# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot

class CFiniteDistribution:
    """This class represents a finite distribution of a continuous random variable 
       by giving a list of values among with weights for the corresponding delta 
       distributions."""
       
    def __init__(self,Values=None,Weights=None,Normalize=True):
        """This constructor initializes this distribution with the given values in 
           the given array and and the given weights. If no weights are given, all 
           values get the same weight and the weights sum to 1.0. If Normalize is 
           True, the weights are scaled to sum to 1.0, otherwise they are used as 
           provided. If Values is None, the distribution is defined as invalid."""
        if(not Values is None):
            self.Init(Values,Weights,Normalize);
    
    def Init(self,Values,Weights=None,Normalize=True):
        """A synonym for the constructor, which can be called at any time."""
        Values=np.asarray(Values).flatten();
        if(Weights is None):
            Weights=np.ones(Values.shape);
        Weights=np.asarray(Weights).flatten();
        if(Normalize):
            Weights/=np.sum(Weights);
        # Sort the input and eliminate duplicate values
        self.Values,WeightIndices=np.unique(Values,False,True);
        self.Weights=np.zeros(self.Values.shape);
        for i,iWeight in enumerate(WeightIndices):
            self.Weights[iWeight]+=Weights[i];

    def InitUnsafe(self,Values,Weights):
        """A faster variant of Init(). Values and Weights have to be given 
           explicitly and values have to be sorted."""
        self.Values=np.asarray(Values).flatten();
        self.Weights=np.asarray(Weights).flatten();

    def GetGraph(self,XMin=0.0,XMax=2.0*np.pi):    
        """This function returns a tuple consisting of arrays for the x-coordinates 
           and y-coordinates of points, which need to be connected in order to 
           obtain a graph of this distribution."""
        nValue=np.size(self.Values);
        X=np.zeros(nValue*2+2);
        Y=np.zeros(nValue*2+2);
        SummedWeight=0.0;
        X[0]=XMin;
        Y[0]=SummedWeight;
        for i in range(nValue):
            X[i*2+1:i*2+3]=self.Values[i];
            Y[i*2+1]=SummedWeight;
            SummedWeight+=self.Weights[i];
            Y[i*2+2]=SummedWeight;
        X[nValue*2+1]=XMax;
        Y[nValue*2+1]=SummedWeight;
        return X,Y;

    def GetCDF(self,X):
        """This function evaluates the CDF of this distribution at the given value 
           and returns the result. X can be an array. In this case an array of 
           identical shape containing all values of the CDF at these samples is 
           returned."""
        if(np.asarray(X).size==1):
            return np.sum(np.where(self.Values<=X,self.Weights,np.zeros(self.Weights.shape)));
        else:
            CDFX,CDFY=self.GetGraph(np.min(X),np.max(X));
            return np.interp(X,CDFX,CDFY);

    def Plot(self,LinePlot=None,XMin=0.0,XMax=2.0*np.pi,**KeywordArguments):
        """This function creates a plot of the cumulative distribution function of 
           this distribution. Keyword arguments are passed on to matplotlib. The 
           return value of pyplot.plot is returned. If you pass the return value of 
           a previous call to LinePlot, this plot is changed to show the new graph 
           and no new plot is generated."""
        X,Y=self.GetGraph(XMin,XMax);
        if(LinePlot):
            return LinePlot[0].set_data(X,Y);
        else:
            return pyplot.plot(X,Y,**KeywordArguments);

    def Integrate(self,Function):
        """Given a function, which can be evaluated on one-dimensional numpy arrays 
           this function returns the result of integral Function(x)dP(x) where P is 
           this distribution."""
        return np.sum(Function(self.Values)*self.Weights);

    def GetNorm(self):
        """This function returns the sum of all weights in this finite distribution. 
           For probability distributions this is 1.0."""
        return self.Integrate(lambda X: 1.0);

    def Normalize(self):
        """This function makes this distribution have norm one, i.e. it turns it 
           into a probability distribution. This is achieved by means of scaling."""
        self.Weights/=self.GetNorm();

    def __rmul__(self,LHS):
        """Multiplies the CDF of this distribution by the given scalar and returns 
           the resulting finite distribution. This scalar multiplication among with 
           __add__() turns probability distributions into a vector space."""
        return CFiniteDistribution(self.Values,self.Weights*LHS,False);

    def __add__(self,RHS):
        """Adds the CDF of the given distribution to the CDF of this distribution 
           and returns the resulting finite distribution. This addition among with 
           __rmul__() turns probability distributions into a vector space."""
        if(self.Values.shape==RHS.Values.shape and (self.Values==RHS.Values).all()):
            Result=CFiniteDistribution();
            Result.InitUnsafe(self.Values,self.Weights+RHS.Weights);
            return Result;
        else:
            return CFiniteDistribution(np.concatenate((self.Values,RHS.Values)),np.concatenate((self.Weights,RHS.Weights)),False);

    ## The values of this distribution in an numpy array. They are sorted.
    Values=None;
    ## The weights for the values in an array of identical shape
    Weights=None;
