# -*- coding: utf-8 -*-
import struct
from os import path
import numpy as np
from scipy import io

## \file
# Provides various functions for convenient reading and writing of image data.


def ArrayToFloatBuffer(Array):
    """This function converts a NumPy array to a binary buffer of single precision 
       values, which can be processed by OpenEXR."""
    return np.asarray(Array).astype(np.float32).tostring();

def FloatBufferToArray(Data,Shape):
    """This function converts a binary buffer of single precision values as expected 
       by OpenEXR to a NumPy array of given Shape."""
    return np.fromstring(Data,dtype=np.float32).reshape(Shape);


def StoreImage(Image,FilePath):
    """Stores an image to a file. It is a wrapper atop wrappers to support PIL image 
       file formats OpenEXR (*.exr) and Matlab data files (*.mat).
      \param Image A real array of shape (Height,Width,3) or (Height,Width). In the 
             first case Image[Y,X,:] defines an RGB triple, in the latter case 
             Image[Y,X] is repeated as brightness for all channels. When writing to 
             PIL images the held values should lie between zero and one.
      \param FilePath The path to an OpenEXR file (*.exr), a Matlab data file 
             (*.mat) or a file with a file format extension supported by PIL. This 
             is where the output image is written to. If it exists already, it is 
             overwritten without prompt. Written OpenEXR files are always single 
             precision RGB images. Written Matlab data files contain a single 
             variable named Image containing the given array.
      \note The function does not perform sRGB conversion. The given values are 
            stored as is for all formats.
      \throws ImportError If PIL or OpenEXR are needed, but not available.
      \sa LoadImage()"""
    # Get the image into shape
    Image=np.asarray(Image);
    if(len(Image.shape)==2):
        Image=Image[:,:,np.newaxis].repeat(3,2);
    # Determine the file type
    Extension=path.splitext(FilePath)[1].lower();
    if(Extension==".exr"):
        import OpenEXR
        import Imath
        # Create and open the output file
        FileHeader=OpenEXR.Header(Image.shape[1],Image.shape[0]);
        FileHeader["channels"]={"R":Imath.Channel(Imath.PixelType(OpenEXR.FLOAT)),"G":Imath.Channel(Imath.PixelType(OpenEXR.FLOAT)),"B":Imath.Channel(Imath.PixelType(OpenEXR.FLOAT))};
        OutputFile=OpenEXR.OutputFile(FilePath,FileHeader);
        # Write the image data
        DataDictionary=dict();
        for i,ChannelName in enumerate(["R","G","B"]):
            DataDictionary[ChannelName]=ArrayToFloatBuffer(Image[:,:,i]);
        OutputFile.writePixels(DataDictionary);
        OutputFile.close();
    elif(Extension==".mat"):
        io.savemat(FilePath,{"Image":Image});
    else:
        from scipy import misc
        # Cast to uint8 because otherwise SciPy does in an undefined manner
        LDRImage=np.asarray(np.clip(np.floor(Image*256.0),0.0,255.0),dtype=np.uint8);
        misc.imsave(FilePath,LDRImage);

def LoadImage(FilePath):
    """This function is the counterpart of StoreImage(). It loads an image from a 
       file and returns it.
      \param FilePath The path to the image file. The choices are the same as for 
             StoreImage().
      \return An array of shape (Height,Width,3) with data type np.float.
      \warning The function is not guaranteed to behave correctly for images which 
               have not been generated in the format produced by StoreImage()."""
    # Determine the file type
    Extension=path.splitext(FilePath)[1].lower();
    if(Extension==".exr"):
        import OpenEXR
        import Imath
        # Load the file
        InputFile=OpenEXR.InputFile(FilePath);
        ChannelData=list();
        Window=InputFile.header()["dataWindow"];
        Shape=(1+Window.max.y-Window.min.y,1+Window.max.x-Window.min.x);
        for ChannelName in ["R","G","B"]:
            ChannelData.append(FloatBufferToArray(InputFile.channel(ChannelName,Imath.PixelType(OpenEXR.FLOAT)),Shape));
        Image=np.dstack(ChannelData);
    elif(Extension==".mat"):
        Image=io.loadmat(FilePath)["Image"];
    else:
        from scipy import misc
        Image=misc.imread(FilePath)*(1.0/255.0);
    return np.asarray(Image,np.float)[:,:,0:3];


def StoreScalarImage(ComplexOrRealImage,FilePath):
    """This function creates a new image file containing the scalar values in the 
       given array.
      \param ComplexOrRealImage A two-dimensional array containing scalar values. 
             The values may be complex or real.
      \param FilePath The path to the image file that is to be created. See 
             StoreImage() for details. The file uses single precision and RGB 
             channels. The red channel contains the real part of the provided 
             values. The green channel contains the imaginary part (if present). 
             Other channels have value zero. To avoid loss of data it is recommended 
             to use OpenEXR files (*.exr) or Matlab data files (*.mat)."""
    if(np.iscomplexobj(ComplexOrRealImage)):
        StoreImage(np.dstack([ComplexOrRealImage.real,ComplexOrRealImage.imag,np.zeros_like(ComplexOrRealImage,dtype=np.float)]),FilePath);
    else:
        StoreImage(ComplexOrRealImage,FilePath);

def LoadScalarImage(FilePath):
    """This function loads an image file containing the scalar values of a two-
       dimensional array as produced by StoreScalarImage().
      \param FilePath The path to an image file as produced by StoreScalarImage().
      \return A two-dimensional np.ndarray as expected by StoreScalarImage()."""
    Image=LoadImage(FilePath);
    if(np.linalg.norm(Image[:,:,1])==0.0):
        return Image[:,:,0];
    else:
        return Image[:,:,0]+1.0j*Image[:,:,1];


def Unpack(Stream,StructFormat):
    """Reads data from a stream using struct.unpack() and returns the resulting 
       tuple. StructFormat is the format string passed to struct.unpack. As many 
       bytes as needed for this format will be read from Stream."""
    StructSize=struct.calcsize(StructFormat);
    Data=Stream.read(StructSize);
    return struct.unpack(StructFormat,Data);
