/*! \file
   This header implements Levinson's algorithm.*/

#include "ComplexAlgebra.fx"

#define TOEPLITZ_SIZE 2
/*! This function solves a special system of linear equations. The equations are 
   defined by a hermitean Toeplitz matrix. The right hand side of the equations is 
   defined by the first canonical basis vector, i.e. the equation defined by the 
   first row has to yield one and all other equations have to yield zero.
  \param OutSolution The solution to the system of linear equations. This is a 
		 column vector holding complex values.
  \param FirstColumn This array has to provide the complex entries of the 
		 first column of the Toeplitz matrix. Since the matrix is hermitean and 
		 Toeplitz all other entries follow. Note that the first entry in this array 
		 must be real, i.e. FirstColumn[0].y==0.0f.
  \note The function implements the Levinson algorithm and takes time 
		O(TOEPLITZ_SIZE*TOEPLITZ_SIZE). */
void SolveCanonicalToeplitzEquation(out float2 OutSolution[TOEPLITZ_SIZE],float2 FirstColumn[TOEPLITZ_SIZE]){
	// We solve the system iteratively for increasing TOEPLITZ_SIZE starting with 
	// TOEPLITZ_SIZE==1
	OutSolution[0]=float2(1.0f/FirstColumn[0].x,0.0f);
	// Proceed for increasingly larger Toeplitz matrices. In every iteration the 
	// solution for TOEPLITZ_SIZE==m is used to construct the solution for 
	// TOEPLITZ_SIZE==m+1.
	[unroll] for(int m=1;m!=TOEPLITZ_SIZE;++m){
		// Evaluate a dot product between the first column of the Toeplitz matrix 
		// and the previous solution
		float2 DotProduct=float2(0.0f,0.0f);
		[unroll] for(int j=0;j!=m;++j){
			DotProduct+=Multiply(OutSolution[j],FirstColumn[m-j]);
		}
		// Turn it into factors for the recursion
		float Factor1=1.0f/(1.0f-dot(DotProduct,DotProduct));
		float2 Factor2=-Factor1*DotProduct;

		// Construct the new solution as linear combination of the previous solution 
		// and a flipped and shifted version of the previous solution. Essentially, 
		// this is very simple but some special considerations are needed to get it 
		// done in-situ.
		float FirstCanonicalLength=OutSolution[0].x;
		[unroll] for(int j=1;2*j<=m;++j){
			float2 PreviousEntry=OutSolution[j];
			float2 PreviousEntryFlipped=OutSolution[m-j];
			OutSolution[j]=Factor1*PreviousEntry+Multiply(Factor2,Conjugate(PreviousEntryFlipped));
			OutSolution[m-j]=Factor1*PreviousEntryFlipped+Multiply(Factor2,Conjugate(PreviousEntry));
		}
		OutSolution[0].x=Factor1*FirstCanonicalLength;
		OutSolution[m]=Factor2*FirstCanonicalLength;
	}
}


#define TOEPLITZ_SIZE 3
void SolveCanonicalToeplitzEquation(out float2 OutSolution[TOEPLITZ_SIZE],float2 FirstColumn[TOEPLITZ_SIZE]){
	// We solve the system iteratively for increasing TOEPLITZ_SIZE starting with 
	// TOEPLITZ_SIZE==1
	OutSolution[0]=float2(1.0f/FirstColumn[0].x,0.0f);
	// Proceed for increasingly larger Toeplitz matrices. In every iteration the 
	// solution for TOEPLITZ_SIZE==m is used to construct the solution for 
	// TOEPLITZ_SIZE==m+1.
	[unroll] for(int m=1;m!=TOEPLITZ_SIZE;++m){
		// Evaluate a dot product between the first column of the Toeplitz matrix 
		// and the previous solution
		float2 DotProduct=float2(0.0f,0.0f);
		[unroll] for(int j=0;j!=m;++j){
			DotProduct+=Multiply(OutSolution[j],FirstColumn[m-j]);
		}
		// Turn it into factors for the recursion
		float Factor1=1.0f/(1.0f-dot(DotProduct,DotProduct));
		float2 Factor2=-Factor1*DotProduct;

		// Construct the new solution as linear combination of the previous solution 
		// and a flipped and shifted version of the previous solution. Essentially, 
		// this is very simple but some special considerations are needed to get it 
		// done in-situ.
		float FirstCanonicalLength=OutSolution[0].x;
		[unroll] for(int j=1;2*j<=m;++j){
			float2 PreviousEntry=OutSolution[j];
			float2 PreviousEntryFlipped=OutSolution[m-j];
			OutSolution[j]=Factor1*PreviousEntry+Multiply(Factor2,Conjugate(PreviousEntryFlipped));
			OutSolution[m-j]=Factor1*PreviousEntryFlipped+Multiply(Factor2,Conjugate(PreviousEntry));
		}
		OutSolution[0].x=Factor1*FirstCanonicalLength;
		OutSolution[m]=Factor2*FirstCanonicalLength;
	}
}


#define TOEPLITZ_SIZE 4
void SolveCanonicalToeplitzEquation(out float2 OutSolution[TOEPLITZ_SIZE],float2 FirstColumn[TOEPLITZ_SIZE]){
	// We solve the system iteratively for increasing TOEPLITZ_SIZE starting with 
	// TOEPLITZ_SIZE==1
	OutSolution[0]=float2(1.0f/FirstColumn[0].x,0.0f);
	// Proceed for increasingly larger Toeplitz matrices. In every iteration the 
	// solution for TOEPLITZ_SIZE==m is used to construct the solution for 
	// TOEPLITZ_SIZE==m+1.
	[unroll] for(int m=1;m!=TOEPLITZ_SIZE;++m){
		// Evaluate a dot product between the first column of the Toeplitz matrix 
		// and the previous solution
		float2 DotProduct=float2(0.0f,0.0f);
		[unroll] for(int j=0;j!=m;++j){
			DotProduct+=Multiply(OutSolution[j],FirstColumn[m-j]);
		}
		// Turn it into factors for the recursion
		float Factor1=1.0f/(1.0f-dot(DotProduct,DotProduct));
		float2 Factor2=-Factor1*DotProduct;

		// Construct the new solution as linear combination of the previous solution 
		// and a flipped and shifted version of the previous solution. Essentially, 
		// this is very simple but some special considerations are needed to get it 
		// done in-situ.
		float FirstCanonicalLength=OutSolution[0].x;
		[unroll] for(int j=1;2*j<=m;++j){
			float2 PreviousEntry=OutSolution[j];
			float2 PreviousEntryFlipped=OutSolution[m-j];
			OutSolution[j]=Factor1*PreviousEntry+Multiply(Factor2,Conjugate(PreviousEntryFlipped));
			OutSolution[m-j]=Factor1*PreviousEntryFlipped+Multiply(Factor2,Conjugate(PreviousEntry));
		}
		OutSolution[0].x=Factor1*FirstCanonicalLength;
		OutSolution[m]=Factor2*FirstCanonicalLength;
	}
}


#define TOEPLITZ_SIZE 5
void SolveCanonicalToeplitzEquation(out float2 OutSolution[TOEPLITZ_SIZE],float2 FirstColumn[TOEPLITZ_SIZE]){
	// We solve the system iteratively for increasing TOEPLITZ_SIZE starting with 
	// TOEPLITZ_SIZE==1
	OutSolution[0]=float2(1.0f/FirstColumn[0].x,0.0f);
	// Proceed for increasingly larger Toeplitz matrices. In every iteration the 
	// solution for TOEPLITZ_SIZE==m is used to construct the solution for 
	// TOEPLITZ_SIZE==m+1.
	[unroll] for(int m=1;m!=TOEPLITZ_SIZE;++m){
		// Evaluate a dot product between the first column of the Toeplitz matrix 
		// and the previous solution
		float2 DotProduct=float2(0.0f,0.0f);
		[unroll] for(int j=0;j!=m;++j){
			DotProduct+=Multiply(OutSolution[j],FirstColumn[m-j]);
		}
		// Turn it into factors for the recursion
		float Factor1=1.0f/(1.0f-dot(DotProduct,DotProduct));
		float2 Factor2=-Factor1*DotProduct;

		// Construct the new solution as linear combination of the previous solution 
		// and a flipped and shifted version of the previous solution. Essentially, 
		// this is very simple but some special considerations are needed to get it 
		// done in-situ.
		float FirstCanonicalLength=OutSolution[0].x;
		[unroll] for(int j=1;2*j<=m;++j){
			float2 PreviousEntry=OutSolution[j];
			float2 PreviousEntryFlipped=OutSolution[m-j];
			OutSolution[j]=Factor1*PreviousEntry+Multiply(Factor2,Conjugate(PreviousEntryFlipped));
			OutSolution[m-j]=Factor1*PreviousEntryFlipped+Multiply(Factor2,Conjugate(PreviousEntry));
		}
		OutSolution[0].x=Factor1*FirstCanonicalLength;
		OutSolution[m]=Factor2*FirstCanonicalLength;
	}
}


#define TOEPLITZ_SIZE 6
void SolveCanonicalToeplitzEquation(out float2 OutSolution[TOEPLITZ_SIZE],float2 FirstColumn[TOEPLITZ_SIZE]){
	// We solve the system iteratively for increasing TOEPLITZ_SIZE starting with 
	// TOEPLITZ_SIZE==1
	OutSolution[0]=float2(1.0f/FirstColumn[0].x,0.0f);
	// Proceed for increasingly larger Toeplitz matrices. In every iteration the 
	// solution for TOEPLITZ_SIZE==m is used to construct the solution for 
	// TOEPLITZ_SIZE==m+1.
	[unroll] for(int m=1;m!=TOEPLITZ_SIZE;++m){
		// Evaluate a dot product between the first column of the Toeplitz matrix 
		// and the previous solution
		float2 DotProduct=float2(0.0f,0.0f);
		[unroll] for(int j=0;j!=m;++j){
			DotProduct+=Multiply(OutSolution[j],FirstColumn[m-j]);
		}
		// Turn it into factors for the recursion
		float Factor1=1.0f/(1.0f-dot(DotProduct,DotProduct));
		float2 Factor2=-Factor1*DotProduct;

		// Construct the new solution as linear combination of the previous solution 
		// and a flipped and shifted version of the previous solution. Essentially, 
		// this is very simple but some special considerations are needed to get it 
		// done in-situ.
		float FirstCanonicalLength=OutSolution[0].x;
		[unroll] for(int j=1;2*j<=m;++j){
			float2 PreviousEntry=OutSolution[j];
			float2 PreviousEntryFlipped=OutSolution[m-j];
			OutSolution[j]=Factor1*PreviousEntry+Multiply(Factor2,Conjugate(PreviousEntryFlipped));
			OutSolution[m-j]=Factor1*PreviousEntryFlipped+Multiply(Factor2,Conjugate(PreviousEntry));
		}
		OutSolution[0].x=Factor1*FirstCanonicalLength;
		OutSolution[m]=Factor2*FirstCanonicalLength;
	}
}


#define TOEPLITZ_SIZE 7
void SolveCanonicalToeplitzEquation(out float2 OutSolution[TOEPLITZ_SIZE],float2 FirstColumn[TOEPLITZ_SIZE]){
	// We solve the system iteratively for increasing TOEPLITZ_SIZE starting with 
	// TOEPLITZ_SIZE==1
	OutSolution[0]=float2(1.0f/FirstColumn[0].x,0.0f);
	// Proceed for increasingly larger Toeplitz matrices. In every iteration the 
	// solution for TOEPLITZ_SIZE==m is used to construct the solution for 
	// TOEPLITZ_SIZE==m+1.
	[unroll] for(int m=1;m!=TOEPLITZ_SIZE;++m){
		// Evaluate a dot product between the first column of the Toeplitz matrix 
		// and the previous solution
		float2 DotProduct=float2(0.0f,0.0f);
		[unroll] for(int j=0;j!=m;++j){
			DotProduct+=Multiply(OutSolution[j],FirstColumn[m-j]);
		}
		// Turn it into factors for the recursion
		float Factor1=1.0f/(1.0f-dot(DotProduct,DotProduct));
		float2 Factor2=-Factor1*DotProduct;

		// Construct the new solution as linear combination of the previous solution 
		// and a flipped and shifted version of the previous solution. Essentially, 
		// this is very simple but some special considerations are needed to get it 
		// done in-situ.
		float FirstCanonicalLength=OutSolution[0].x;
		[unroll] for(int j=1;2*j<=m;++j){
			float2 PreviousEntry=OutSolution[j];
			float2 PreviousEntryFlipped=OutSolution[m-j];
			OutSolution[j]=Factor1*PreviousEntry+Multiply(Factor2,Conjugate(PreviousEntryFlipped));
			OutSolution[m-j]=Factor1*PreviousEntryFlipped+Multiply(Factor2,Conjugate(PreviousEntry));
		}
		OutSolution[0].x=Factor1*FirstCanonicalLength;
		OutSolution[m]=Factor2*FirstCanonicalLength;
	}
}


#define TOEPLITZ_SIZE 8
void SolveCanonicalToeplitzEquation(out float2 OutSolution[TOEPLITZ_SIZE],float2 FirstColumn[TOEPLITZ_SIZE]){
	// We solve the system iteratively for increasing TOEPLITZ_SIZE starting with 
	// TOEPLITZ_SIZE==1
	OutSolution[0]=float2(1.0f/FirstColumn[0].x,0.0f);
	// Proceed for increasingly larger Toeplitz matrices. In every iteration the 
	// solution for TOEPLITZ_SIZE==m is used to construct the solution for 
	// TOEPLITZ_SIZE==m+1.
	[unroll] for(int m=1;m!=TOEPLITZ_SIZE;++m){
		// Evaluate a dot product between the first column of the Toeplitz matrix 
		// and the previous solution
		float2 DotProduct=float2(0.0f,0.0f);
		[unroll] for(int j=0;j!=m;++j){
			DotProduct+=Multiply(OutSolution[j],FirstColumn[m-j]);
		}
		// Turn it into factors for the recursion
		float Factor1=1.0f/(1.0f-dot(DotProduct,DotProduct));
		float2 Factor2=-Factor1*DotProduct;

		// Construct the new solution as linear combination of the previous solution 
		// and a flipped and shifted version of the previous solution. Essentially, 
		// this is very simple but some special considerations are needed to get it 
		// done in-situ.
		float FirstCanonicalLength=OutSolution[0].x;
		[unroll] for(int j=1;2*j<=m;++j){
			float2 PreviousEntry=OutSolution[j];
			float2 PreviousEntryFlipped=OutSolution[m-j];
			OutSolution[j]=Factor1*PreviousEntry+Multiply(Factor2,Conjugate(PreviousEntryFlipped));
			OutSolution[m-j]=Factor1*PreviousEntryFlipped+Multiply(Factor2,Conjugate(PreviousEntry));
		}
		OutSolution[0].x=Factor1*FirstCanonicalLength;
		OutSolution[m]=Factor2*FirstCanonicalLength;
	}
}


#define TOEPLITZ_SIZE 9
void SolveCanonicalToeplitzEquation(out float2 OutSolution[TOEPLITZ_SIZE],float2 FirstColumn[TOEPLITZ_SIZE]){
	// We solve the system iteratively for increasing TOEPLITZ_SIZE starting with 
	// TOEPLITZ_SIZE==1
	OutSolution[0]=float2(1.0f/FirstColumn[0].x,0.0f);
	// Proceed for increasingly larger Toeplitz matrices. In every iteration the 
	// solution for TOEPLITZ_SIZE==m is used to construct the solution for 
	// TOEPLITZ_SIZE==m+1.
	[unroll] for(int m=1;m!=TOEPLITZ_SIZE;++m){
		// Evaluate a dot product between the first column of the Toeplitz matrix 
		// and the previous solution
		float2 DotProduct=float2(0.0f,0.0f);
		[unroll] for(int j=0;j!=m;++j){
			DotProduct+=Multiply(OutSolution[j],FirstColumn[m-j]);
		}
		// Turn it into factors for the recursion
		float Factor1=1.0f/(1.0f-dot(DotProduct,DotProduct));
		float2 Factor2=-Factor1*DotProduct;

		// Construct the new solution as linear combination of the previous solution 
		// and a flipped and shifted version of the previous solution. Essentially, 
		// this is very simple but some special considerations are needed to get it 
		// done in-situ.
		float FirstCanonicalLength=OutSolution[0].x;
		[unroll] for(int j=1;2*j<=m;++j){
			float2 PreviousEntry=OutSolution[j];
			float2 PreviousEntryFlipped=OutSolution[m-j];
			OutSolution[j]=Factor1*PreviousEntry+Multiply(Factor2,Conjugate(PreviousEntryFlipped));
			OutSolution[m-j]=Factor1*PreviousEntryFlipped+Multiply(Factor2,Conjugate(PreviousEntry));
		}
		OutSolution[0].x=Factor1*FirstCanonicalLength;
		OutSolution[m]=Factor2*FirstCanonicalLength;
	}
}
