# -*- coding: utf-8 -*-
import re
from os import path,listdir
import numpy as np
from scipy import ndimage
from matplotlib import pyplot
from InputOutput import Unpack
import TrigonometricMomentImageStack
from Plotting import GetSubplotLayout

# The tolerance for errors in frequencies in Hz
FrequencyTolerance=10.0;

class CMomentImage:
    """This class encapsulates a single real or complex moment image among with 
       meta-data."""
    
    def __init__(self,Image,ExposureTime,Frequency,PhaseShift,TimeStamp=0.0,Tap=0,Take=0):
        """Initializes with the given attributes."""
        self.Image=np.asarray(Image);
        self.ExposureTime=ExposureTime;
        self.Frequency=Frequency;
        self.PhaseShift=PhaseShift;
        self.TimeStamp=TimeStamp;
        self.Tap=Tap;
        self.Take=Take;
    
    def GetLinearImage(self,ReferenceExposureTime=1.92e-3):
        """Computes an estimate of an image with the given exposure time (in 
           seconds) using the information in this image and returns it."""
        return self.Image*ReferenceExposureTime/self.ExposureTime;
    
    def GetWidth(self):
        """Returns the width of this image in pixels."""
        return self.Image.shape[1];

    def GetHeight(self):
        """Returns the height of this image in pixels."""
        return self.Image.shape[0];
    
    def GetColor(self):
        """For plots it can be useful to visualize different taps with different 
           colors. Therefore, this function returns a color that identifies the tap 
           index of this moment image."""
        return ["b","g","r","y"][self.Tap%4];
    
    ## A 2D array containing the data of the measurement
    Image=None;
    ## The exposure time of the moment image in seconds
    ExposureTime=0.0;
    ## The frequency used for modulation in Hz
    Frequency=0.0;
    ## The phase shift of the two modulation signals 
    PhaseShift=0.0;
    ## The time stamp of the image. This is a floating point quantity in seconds 
    # with respect to some absolute time that is only fixed per measurement. If 
    # unavailable, it is zero.
    TimeStamp=0.0;
    ## The index of the tap that lead to this moment image. Typically, four moment 
    # images with pi/2 phase shifts are captured in a row and this index provides 
    # the index within this series of measurements. Note that the phase shift 
    # already includes the offset due to the tap index. If multiple moment images 
    # were combined to construct this moment image, this index is the minimum of the 
    # involved indices.
    Tap=0;
    ## The take in which this moment image was recorded. Multiple takes are either 
    # used to average out noise or to record videos. In the latter case this is the 
    # index of the frame.
    Take=0;


class CMeasurement:
    """This class encapsulates a series of measurements consisting of multiple 
       moment images. This is typically more or less unprocessed raw-data from a 
       photonic mixer device."""
    
    def __init__(self):
        """Initializes this series of measurements as empty."""
        self.MomentImageList=list();
        
    def AddCaptureData(self,Path,FilterFunction=lambda ExposureTime,Frequency,PhaseShift,Take: True,SubtractImages=True,IgnoreTaps=False):
        """Loads capture data stored in the given directory using a binary format. 
          \param Path The path of the directory containing the data. The filenames 
                 have to match the default scheme, otherwise they are ignored.
          \param FilterFunction A function determining whether a file is loaded or 
                 not based on the meta-information in the file name. The parameters 
                 are ExposureTime (in seconds), Frequency (in Hertz), PhaseShift (in 
                 radians) and Take (zero-based index). The function should return 
                 True if the file is to be loaded and False otherwise.
          \param SubtractImages If this is True, pairs of images with with a 
                 relative phase shift of pi are subtracted from one-another before 
                 storing them. Otherwise two separate moment images are stored.
          \param IgnoreTaps If this boolean is True, the function only loads the 
                 first moment image from each file. This can be useful because the 
                 power of the laser diodes drops off during the capture of four 
                 images. This setting overrides SubtractImages."""
        global FrequencyTolerance;
        FileNameRE=r".*?\s+Int\s+([0-9\.]+)\s+Freq\s+([0-9\.]+)\s+Phase\s+([+\-0-9\.]+)\s+Take\s+([0-9]+)\.bin";
        # Create a catalog of available frequencies to assign colors
        FrequencyList=list();
        for FileName in listdir(Path):
            FileNameMatch=re.match(FileNameRE,FileName);
            if(not FileNameMatch is None):
                FrequencyList.append(float(FileNameMatch.group(2))*1.0e6);
        FrequencyList=np.array(sorted(list(frozenset(FrequencyList))));
        # Do the actual loading of images
        for FileName in listdir(Path):
            FilePath=path.join(Path,FileName);
            # Parse the file name
            FileNameMatch=re.match(FileNameRE,FileName);
            if(FileNameMatch==None):
                continue;
            NewExposureTime=float(FileNameMatch.group(1))*1.0e-6;
            NewFrequency=float(FileNameMatch.group(2))*1.0e6;
            NewPhaseShift=float(FileNameMatch.group(3));
            NewTake=int(FileNameMatch.group(4));
            # Determine whether the measurement is relevant
            if(not FilterFunction(NewExposureTime,NewFrequency,NewPhaseShift,NewTake)):
                continue;
            # Read and validate the file header
            File=open(FilePath,"rb");
            CamRec="".join(Unpack(File,"8c"));
            if(CamRec!="Cam Rec\n"):
                print("The image "+FileName+" has an invalid magic value.");
                continue;
            Version,=Unpack(File,"I");
            if(Version!=3):
                print("Unexpected file format version: "+str(Version));
                continue;
            Width,Height=Unpack(File,"II");
            # Load the sequence of phase images
            Img="".join(Unpack(File,"4c"));
            if(Img!="Img\n"):
                print("Invalid image header in "+FileName+".");
            Unpack(File,"i");
            TimeStamp,nPhaseImage=Unpack(File,"QH");
            Unpack(File,"h");
            Unpack(File,"i");
            TimeStamp*=1.0e-6;
            ImageDataLayout="f"*(Width*Height);
            DataOffset=((-(2.0**15)) if nPhaseImage>=4 else 0.0);
            # Each sub image has its own header followed by its raw data
            ImageList=list();
            for i in range(nPhaseImage):
                ReadFrequency,ReadExposureTime,ReadPhase=Unpack(File,"fff");
                ImageList.append(((ReadFrequency*1.0e6,ReadExposureTime*1.0e-6,ReadPhase),np.reshape(Unpack(File,ImageDataLayout),(Height,Width))+DataOffset));
            if(nPhaseImage==5):
                # There is no useful data in the first phase image if five are 
                # stored
                ImageList=ImageList[1:];
                nPhaseImage=4;
            # Only store a single image if requested
            if(IgnoreTaps):
                OnlySubImage=ImageList[0];
                self.MomentImageList.append(CMomentImage(OnlySubImage[1],OnlySubImage[0][1],OnlySubImage[0][0],OnlySubImage[0][2],TimeStamp,0,NewTake));
            else:
                if(nPhaseImage==4):
                    # Create a cosine image and a sine image
                    if(SubtractImages):
                        CosineImage=ImageList[0][1]-ImageList[2][1];
                        HeldFrequency,HeldExposureTime,HeldPhase=ImageList[0][0];
                        self.MomentImageList.append(CMomentImage(CosineImage,HeldExposureTime,HeldFrequency,HeldPhase,TimeStamp,0,NewTake));
                        SineImage=ImageList[1][1]-ImageList[3][1];
                        HeldFrequency,HeldExposureTime,HeldPhase=ImageList[1][0];
                        self.MomentImageList.append(CMomentImage(SineImage,HeldExposureTime,HeldFrequency,HeldPhase,TimeStamp,1,NewTake));
                    # Store each image individually
                    else:
                        for i,Image in enumerate(ImageList):
                            MomentImage=CMomentImage(Image[1],Image[0][1],Image[0][0],Image[0][2],TimeStamp,i,NewTake);
                            self.MomentImageList.append(MomentImage);
                elif(nPhaseImage==2):
                    # Store the available sine and cosine image
                    CosineImage=ImageList[0];
                    self.MomentImageList.append(CMomentImage(CosineImage[1],CosineImage[0][1],CosineImage[0][0],CosineImage[0][2],TimeStamp,0,NewTake));
                    SineImage=ImageList[1];
                    self.MomentImageList.append(CMomentImage(SineImage[1],SineImage[0][1],SineImage[0][0],SineImage[0][2],TimeStamp,1,NewTake));
                else:
                    print("The file "+FileName+" holds "+str(nPhaseImage)+" images per measurement. This cannot be handled.");
    
    def GetPhaseSet(self,Frequency=None):
        """This function returns a set containing all phases (in radians) for which 
           measurements are available at the given frequency (in Hz). If Frequency 
           is None, phases of all frequencies are used."""
        global FrequencyTolerance;
        if(Frequency is None):
            return frozenset([MomentImage.PhaseShift for MomentImage in self.MomentImageList]);
        else:
            return frozenset([MomentImage.PhaseShift for MomentImage in self.MomentImageList if np.abs(MomentImage.Frequency-Frequency)<FrequencyTolerance]);
    
    def RemovePhases(self,PhaseSet,Frequency=None,Tolerance=1.0e-5):
        """This function removes all moment images whose phase shift (in radians) is 
           not found in the given iterable (with some tolerance). If a frequency is 
           given, the operation only removes moment images recorded with this 
           modulation frequency (in Hz)."""
        global FrequencyTolerance;
        PhaseArray=np.exp(1.0j*np.asarray(list(PhaseSet)));
        if(not Frequency is None):
            self.MomentImageList=[MomentImage for MomentImage in self.MomentImageList if np.abs(MomentImage.Frequency-Frequency)>FrequencyTolerance or np.min(np.abs(PhaseArray-np.exp(1.0j*MomentImage.PhaseShift)))<Tolerance];
        else:
            self.MomentImageList=[MomentImage for MomentImage in self.MomentImageList if np.min(np.abs(PhaseArray-np.exp(1.0j*MomentImage.PhaseShift)))<Tolerance];
            
    def GetSortedTakes(self):
        """This function constructs and returns a dictionary containing all moment 
           images in this measurement. The keys have the form (ExposureTime,
           Frequency,PhaseShift,Tap) the values are lists of indices into 
           self.MomentImageList. The lists are sorted by the index of the take."""
        Result=dict();
        for i,MomentImage in enumerate(self.MomentImageList):
            Key=(MomentImage.ExposureTime,MomentImage.Frequency,MomentImage.PhaseShift,MomentImage.Tap);
            if(Key not in Result):
                Result[Key]=list();
            Result[Key].append(i);
        for IndexList in Result.values():
            IndexList.sort(key=lambda Index: self.MomentImageList[Index].Take);
        return Result;

    def FixVideoOutliers(self,CurvatureThreshold=20.0,DerivativeThreshold=15.0,EnablePlotting=False):
        """This function assumes that this measurement provides data for a video 
           where each frame (i.e. take) holds the same set of images captured at 
           different times. It then searches for images, which are statistical 
           outliers as a whole (our prototype hardware produces such outliers). All 
           found outliers are replaced by corresponding data from the previous 
           frame.
          \param CurvatureThreshold,DerivativeThreshold Thresholds on the discrete 
                 curvature and divided difference of the median of the images over 
                 time. Greater thresholds discard less data.
          \param EnablePlotting If this boolean is True, a matplotlib figure is 
                 created to visualize which images have been marked as outlier."""
        MomentImageDict=self.GetSortedTakes();
        if(EnablePlotting):
            Figure=pyplot.figure();
            Figure.suptitle("Outlier detection");
            nRow,nColumn=GetSubplotLayout(len(MomentImageDict));
            iPlot=1;
        for Key,IndexList in MomentImageDict.items():
            # Compute medians per image
            Medians=np.asarray([np.median(self.MomentImageList[i].Image) for i in IndexList]);
            # Compute discrete curvature
            Curvatures=np.abs(Medians-0.5*(np.roll(Medians,-1)+np.roll(Medians,1)));
            # Compute left and right derivatives
            LeftDerivatives=Medians-np.roll(Medians,-1);
            RightDerivatives=np.roll(Medians,1)-Medians;
            # The sign of the derivatives must be opposite and the thresholds have 
            # to be surpassed
            IsOutlier=np.logical_and(LeftDerivatives*RightDerivatives<0.0,np.logical_and(np.abs(Curvatures)>CurvatureThreshold,np.logical_and(np.abs(LeftDerivatives)>DerivativeThreshold,np.abs(RightDerivatives)>DerivativeThreshold)));
            # If three points in a row have been detected as outliers the one in the 
            # middle is probably not an outlier
            IsOutlier=np.logical_and(IsOutlier,np.logical_or(np.logical_not(np.roll(IsOutlier,-1)),np.logical_not(np.roll(IsOutlier,1)))); 
            # Replace outliers
            for i,Index in enumerate(IndexList):
                if(IsOutlier[i]):
                    PreviousIndex=IndexList[i-1] if i>0 else IndexList[len(IndexList)-1];
                    self.MomentImageList[Index].Image=self.MomentImageList[PreviousIndex].Image;
            # Plot the result
            if(EnablePlotting):
                NewMedians=np.asarray([np.median(self.MomentImageList[i].Image) for i in IndexList]);
                Axes=Figure.add_subplot(nRow,nColumn,iPlot);
                iPlot+=1;
                Axes.set_title("%.0fus, %.0fMHz, %.2f, %d"%(Key[0]*1.0e6,Key[1]*1.0e-6,Key[2],Key[3]));
                pyplot.plot(Medians);
                pyplot.plot(NewMedians);
                for i,Outlier in enumerate(IsOutlier):
                    if(Outlier):
                        pyplot.axvline(i);

    def FixStaticOutliers(self,MinFractionToKeep=0.9,EnablePlotting=False):
        """This function assumes that this measurement provides data of a static 
           scene captured with many takes. It then searches for images, which are 
           statistical outliers as a whole (when arccos phase sampling is enabled 
           our prototype hardware produces such outliers). All found outliers are 
           replaced by the mean of non-outliers.
          \param MinFractionToKeep This fraction defines the minimal amount of 
                 frames to keep (per frequency, take, tap and exposure).
          \param EnablePlotting If this boolean is True, a matplotlib figure is 
                 created to visualize which images have been marked as outlier."""
        MomentImageDict=self.GetSortedTakes();
        if(EnablePlotting):
            Figure=pyplot.figure();
            Figure.suptitle("Outlier detection");
            nRow,nColumn=GetSubplotLayout(len(MomentImageDict));
            iPlot=1;
        nImage=0;
        nOutlier=0;
        for Key,IndexList in MomentImageDict.items():
            # Compute medians per image
            Medians=np.asarray([np.median(self.MomentImageList[i].Image) for i in IndexList]);
            # Medians should not be far from the median of medians
            MedianOfMedians=np.median(Medians);
            Features=np.abs(Medians-MedianOfMedians);
            # Determine the threshold for outliers
            MinThreshold=np.percentile(Features,MinFractionToKeep*100.0);
            Threshold=MinThreshold*0.8+np.max(Features)*0.2;
            # Identify outliers
            IsOutlier=Features>Threshold;
            # Replace outliers by the mean of non-outliers
            MeanImage=np.mean(np.dstack([self.MomentImageList[Index].Image for i,Index in enumerate(IndexList) if not IsOutlier[i]]),2);
            for i,Index in enumerate(IndexList):
                if(IsOutlier[i]):
                    self.MomentImageList[Index].Image=MeanImage;
            # Update counts
            nOutlier+=np.count_nonzero(IsOutlier);
            nImage+=np.size(IsOutlier);
            # Plot the result
            if(EnablePlotting):
                NewMedians=np.asarray([np.median(self.MomentImageList[i].Image) for i in IndexList]);
                Axes=Figure.add_subplot(nRow,nColumn,iPlot);
                iPlot+=1;
                Axes.set_title("%.0fus, %.0fMHz, %.2f, %d"%(Key[0]*1.0e6,Key[1]*1.0e-6,Key[2],Key[3]));
                pyplot.plot(Medians);
                pyplot.plot(NewMedians);
                for i,Outlier in enumerate(IsOutlier):
                    if(Outlier):
                        pyplot.axvline(i);
        print("%.2f%% of all images have been recognized as outliers."%((100.0*nOutlier)/nImage));

    def GetTrigonometricMomentImage(self,Frequency,Take=None):
        """This function uses the available measurements to construct an image of 
           trigonometric moments for the given frequency. The image is returned. If 
           no measurements or too few measurements are available for the given 
           frequency, it returns None. The complex scaling of the returned image is 
           somewhat arbitrary due to lack of knowledge of the modulation function. 
           If Take is not None only moment images for this take are considered."""
        global FrequencyTolerance;
        # Get all relevant moment images
        MomentImageSubList=list();
        for MomentImage in self.MomentImageList:
            if(np.abs(MomentImage.Frequency-Frequency)<FrequencyTolerance and (Take==None or MomentImage.Take==Take)):
                MomentImageSubList.append(MomentImage);
        if(len(MomentImageSubList)==0):
            return None;
        # Construct the output moment image
        TrigonometricMomentImage=np.zeros(MomentImageSubList[0].Image.shape[0:2],dtype=np.complex);
        for MomentImage in MomentImageSubList:
            TrigonometricMomentImage+=np.exp(1.0j*MomentImage.PhaseShift)*MomentImage.GetLinearImage();
        TrigonometricMomentImage/=len(MomentImageSubList);
        return TrigonometricMomentImage;
        
    def GetTrigonometricMomentImageStack(self,FrequencyList,Take=None):
        """Invokes GetTrigonometricMomentImage() repeatedly for the given 
           frequencies to construct a CTrigonometricMomentImageStack matching this 
           measurement. The result is returned. If required information is missing 
           a ValueError is thrown. If data is lacking for frequency zero, the 
           function generates an all zero moment image to ensure that biasing 
           generates a meaningful zeroth moment everywhere. If Take is not None only 
           moment images for this take are considered."""
        MomentImageList=list();
        for Frequency in FrequencyList:
            # Generate the moment image
            MomentImage=self.GetTrigonometricMomentImage(Frequency,Take);
            if(MomentImage is None):
                if(Frequency!=0.0 or len(self.MomentImageList)==0):
                    raise ValueError("Lacking information for frequency "+str(Frequency)+". No moment image has been generated.");
                else:
                    # The zeroth moment image can be estimated through biasing. 
                    # Producing an all zero moment image ensures that this happens 
                    # for all pixels.
                    MomentImage=np.zeros_like(self.MomentImageList[0].Image,dtype=np.float);
            # Store the image
            MomentImageList.append(MomentImage);
        return TrigonometricMomentImageStack.CTrigonometricMomentImageStack(MomentImageList);

    def Smooth(self,StandardDeviationInPixels):
        """Applies smoothing to all held moment images using a Gaussian filter with 
           the specified standard deviation in pixels."""
        if(StandardDeviationInPixels>0.0):
            for MomentImage in self.MomentImageList:
                MomentImage.Image=ndimage.filters.gaussian_filter(MomentImage.Image,StandardDeviationInPixels);

    def PlotValueRanges(self,FrequencyList):
        """This function considers the moment images for the frequencies in the 
           given list and visualizes their minimal and maximal values through 
           scatter plots using the phase as x-axis."""
        global FrequencyTolerance;
        Figure=pyplot.figure("Value ranges");
        Figure.suptitle("Minimal and maximal values");
        for i,Frequency in enumerate(FrequencyList):
            Axes=pyplot.subplot(1,len(FrequencyList),i+1);
            Axes.set_title(str(Frequency*1.0e-6)+"Mhz");
            PhaseMinMaxAverageList=list();
            for MomentImage in self.MomentImageList:
                if(np.abs(MomentImage.Frequency-Frequency)<FrequencyTolerance):
                    PhaseMinMaxAverageList.append((MomentImage.PhaseShift,np.percentile(MomentImage.Image,1),np.percentile(MomentImage.Image,99),np.abs(MomentImage.Image).mean()));
            if(len(PhaseMinMaxAverageList)==0):
                continue;
            PhaseMinMaxAverage=np.asarray(PhaseMinMaxAverageList);
            pyplot.scatter(PhaseMinMaxAverage[:,0],PhaseMinMaxAverage[:,1],marker="v");
            pyplot.scatter(PhaseMinMaxAverage[:,0],PhaseMinMaxAverage[:,3],marker=".");
            pyplot.scatter(PhaseMinMaxAverage[:,0],-PhaseMinMaxAverage[:,3],marker=".");
            pyplot.scatter(PhaseMinMaxAverage[:,0],PhaseMinMaxAverage[:,2],marker="^");
        
    def PlotPixelData(self,X,Y,FrequencyList,ShowScatterPlots=True,ShowGraphPlots=False,ShowFourierPlots=False,ShowExposurePlots=False,FigureTitleSuffix=""):
        """This function visualizes information available for the pixel with the 
           given coordinates. In particular, it visualizes how trigonometric moments 
           are constructed from the available information and how these moments are 
           located in relation to each other. It constructs moments based upon the 
           frequencies in the given list. If you pass None for the pixel coordinates 
           the respective coordinates are picked uniformly at random. The boolean 
           parameters define what kind of plots is shown. The given suffix is used 
           in figure titles."""
        global FrequencyTolerance;
        if(X is None):
            X=np.random.randint(self.MomentImageList[0].GetWidth());
        if(Y is None):
            Y=np.random.randint(self.MomentImageList[0].GetHeight());
        # Plot information for all frequencies
        MomentList=list();
        nPlotRow,nPlotColumn=GetSubplotLayout(len(FrequencyList));
        FullTitleSuffix=(" ("+str(FigureTitleSuffix)+")") if len(FigureTitleSuffix)>0 else "";
        if(ShowScatterPlots):
            ScatterFigure=pyplot.figure("Scatter for "+str(X)+","+str(Y)+FullTitleSuffix);
            ScatterFigure.suptitle("Construction of moments for pixel "+str(X)+","+str(Y)+FullTitleSuffix);
        if(ShowGraphPlots):
            GraphFigure=pyplot.figure("Measurements for "+str(X)+","+str(Y)+FullTitleSuffix);
            GraphFigure.suptitle("Measurements for pixel "+str(X)+","+str(Y)+FullTitleSuffix);
        if(ShowFourierPlots):
            FourierFigure=pyplot.figure("Fourier for "+str(X)+","+str(Y)+FullTitleSuffix);
            FourierFigure.suptitle("Magnitudes of Fourier coefficients for pixel "+str(X)+","+str(Y)+FullTitleSuffix);
        if(ShowExposurePlots):
            ExposureFigure=pyplot.figure("Exposure for "+str(X)+","+str(Y)+FullTitleSuffix);
            ExposureFigure.suptitle("Values plotted against exposure for pixel "+str(X)+","+str(Y)+FullTitleSuffix);
        for iPlot,Frequency in enumerate(FrequencyList):
            # Get all relevant moment images
            MomentImageSubList=list();
            for MomentImage in self.MomentImageList:
                if(np.abs(MomentImage.Frequency-Frequency)<FrequencyTolerance):
                    MomentImageSubList.append(MomentImage);
            if(len(MomentImageSubList)==0):
                print("Lacking images for frequency "+str(Frequency)+".");
                continue;
            # Sort the moment images to get nicer plots
            MomentImageSubList.sort(key=lambda MomentImage: (MomentImage.Tap,MomentImage.PhaseShift,MomentImage.ExposureTime));
            if(ShowScatterPlots):
                pyplot.figure(ScatterFigure.number);
                # Gather the points that are averaged together to compute a moment
                MassPoint=np.zeros((len(MomentImageSubList),),dtype=np.complex);
                MassPointColor=list();
                for i,MomentImage in enumerate(MomentImageSubList):
                    MassPoint[i]=MomentImage.GetLinearImage()[Y,X]*np.exp(1.0j*MomentImage.PhaseShift);
                    MassPointColor.append(MomentImage.GetColor());
                # Create aspect preserving axes
                Axes=ScatterFigure.add_subplot(nPlotRow,nPlotColumn,iPlot+1);
                Axes.set_title(str(Frequency*1.0e-6)+"Mhz");
                Axes.set_aspect(1.0);
                Axes.set_xlabel("Real");
                Axes.set_ylabel("Imaginary");
                # Plot the mass points with appropriate colors
                pyplot.scatter(MassPoint.real,MassPoint.imag,c=MassPointColor,linewidth=0.0);
                # Plot lines connecting mass points to the origin (for clarity of 
                # the visualization)
                MassPointToZero=np.vstack([MassPoint[np.newaxis,:],np.zeros((1,MassPoint.shape[0]))]);
                pyplot.plot(MassPointToZero.real,MassPointToZero.imag,color="k",alpha=0.05);
                # Connect points with lines to indicate their order
                pyplot.plot(MassPoint.real,MassPoint.imag,color="k",alpha=0.05);
                # Plot the final moment and the approximation obtained with a single 
                # phase sample
                MomentList.append(MassPoint.mean());
                pyplot.scatter(MassPoint.mean().real,MassPoint.mean().imag,color="k");
            if(ShowExposurePlots):
                pyplot.figure(ExposureFigure.number);
                # Create axes
                Axes=ExposureFigure.add_subplot(nPlotRow,nPlotColumn,iPlot+1);
                Axes.set_title(str(Frequency*1.0e-6)+"Mhz");
                Axes.set_xlabel("Exposure");
                Axes.set_ylabel("Measurement");
                # Plot the data
                ExposureMeasurement=np.zeros((len(MomentImageSubList),2));
                ColorList=list();
                for i,MomentImage in enumerate(MomentImageSubList):
                    ExposureMeasurement[i,0]=MomentImage.ExposureTime;
                    ExposureMeasurement[i,1]=MomentImage.Image[Y,X];
                    ColorList.append(MomentImage.GetColor());
                pyplot.plot(ExposureMeasurement[:,0],ExposureMeasurement[:,1],"-o",color="k",alpha=0.1);
                pyplot.scatter(ExposureMeasurement[:,0],ExposureMeasurement[:,1],c=ColorList);
                pyplot.xlim(0.0,np.max(ExposureMeasurement[:,0])*1.1);
                pyplot.scatter([0.0],[0.0],marker="+");
            # Plot the available data as function of the phase
            PhaseValue=np.zeros((len(MomentImageSubList),2));
            PhaseValueColor=list();
            for i,MomentImage in enumerate(MomentImageSubList):
                PhaseValue[i,0]=MomentImage.PhaseShift;
                PhaseValue[i,1]=MomentImage.GetLinearImage()[Y,X];
                PhaseValueColor.append(MomentImage.GetColor());
            PhaseValue[:,0]=np.modf(20.0*np.pi+PhaseValue[:,0]/(2.0*np.pi))[0]*2.0*np.pi;
            if(ShowGraphPlots):
                pyplot.figure(GraphFigure.number);
                Axes=GraphFigure.add_subplot(nPlotRow,nPlotColumn,iPlot+1);
                Axes.set_title(str(Frequency*1.0e-6)+"Mhz");
                Axes.set_xlabel("Phase");
                Axes.set_ylabel("Value");
                pyplot.scatter(PhaseValue[:,0],PhaseValue[:,1],c=PhaseValueColor,linewidth=0.0);
                Axes.set_xlim(0.0,2.0*np.pi);
                Axes.set_xticks(np.linspace(0.0,2.0*np.pi,5));
                Axes.set_xticklabels(["0",r"$\frac{\pi}{2}$",r"$\pi$",r"$\frac{3\cdot\pi}{2}$",r"$2\cdot\pi$"]);
            if(ShowFourierPlots):
                # Compute Fourier coefficients independent of the phase sampling by 
                # fitting sinusoidals
                nFourierCoefficient=min(30,len(frozenset(PhaseValue[:,0])));
                Vandermonde=np.exp(1.0j*PhaseValue[:,0])[:,np.newaxis].repeat(nFourierCoefficient,1)**np.arange(nFourierCoefficient)[np.newaxis,:].repeat(PhaseValue.shape[0],0);
                FourierCoefficient=np.linalg.lstsq(Vandermonde,PhaseValue[:,1])[0];
                # Plot the magnitudes of the Fourier coefficients
                pyplot.figure(FourierFigure.number);
                Axes=FourierFigure.add_subplot(nPlotRow,nPlotColumn,iPlot+1);
                Axes.set_title(str(Frequency*1.0e-6)+"Mhz");
                Axes.set_xlabel("Frequency");
                Axes.set_ylabel("Magnitude");
                pyplot.bar(np.asarray(range(nFourierCoefficient))+0.1,np.abs(FourierCoefficient));
                Axes.set_xlim(0,nFourierCoefficient);

    def ShowMomentImages(self,FrequencyList):
        """This function displays available moment images for the given frequencies 
           in unmodified form. It is particularly useful for identifying over- or 
           under exposed regions."""
        global FrequencyTolerance;
        # Gather and count the relevant images
        FrequencyMomentImageList=list();
        nImage=0;
        for Frequency in FrequencyList:
            MomentImageSubList=list();
            for MomentImage in self.MomentImageList:
                if(np.abs(MomentImage.Frequency-Frequency)<FrequencyTolerance):
                    MomentImageSubList.append(MomentImage);
                    nImage+=1;
            FrequencyMomentImageList.append(MomentImageSubList);
        # Display the images
        Figure=pyplot.figure("Measured moment images");
        Figure.suptitle("Measured moment images");
        FontDict={"fontsize":10.0,"fontweight":1.0,"verticalalignment":"baseline","horizontalalignment":"center"};
        nRow,nColumn=GetSubplotLayout(nImage);
        iImage=0;
        for MomentImageSubList in FrequencyMomentImageList:
            for MomentImage in MomentImageSubList:
                Axes=pyplot.subplot(nRow,nColumn,iImage+1,xticks=[],yticks=[]);
                iImage+=1;
                Axes.set_title(str(MomentImage.Frequency*1.0e-6)+r"Mhz, $\varphi$="+str(np.round(MomentImage.PhaseShift*180.0/np.pi,1))+r"$\degree$",FontDict);
                pyplot.imshow(MomentImage.Image);
                #pyplot.colorbar();

    ## A list of CMomentImage objects providing all data for this measurement
    MomentImageList=list();
