# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot
from FiniteDistribution import CFiniteDistribution
from Measurement import CMeasurement,FrequencyTolerance

def ProcessMeasurement(MeasurementDirectory,FrequencyList,CalibrationDirectory,ReferenceDistribution=CFiniteDistribution([0.0]),OutputDirectory="MomentImages",**KeywordArguments):
    """This function processes the specified measurement generating trigonometric 
       moment images for the given frequencies and plotting analysis information.
      \param MeasurementDirectory All selected files in this directory are loaded as 
             available data for the measurement.
      \param FrequencyList An iterable of modulation frequencies in Hz (not MHz). 
             The function attempts to produce moment images for each of these 
             frequencies. If data is missing for one of these frequencies, the 
             entire function fails. For biasing it is assumed, that 
             FrequencyList[i]==i*FrequencyList[1].
      \param CalibrationDirectory It is assumed that the files in the given 
             directory contain measurements taken with the same configuration as the 
             measurement but on a different scene. The scene should produce the 
             impulse response given by ReferenceDistribution on every pixel. This is 
             typically achieved by pointing the light source directly at the sensor 
             without lens and passing a Dirac-delta for ReferenceDistribution.
      \param ReferenceDistribution This parameter is forwarded to 
             CTrigonometricMomentImageStack.GetPerPixelCalibration().
      \param OutputDirectory The path to a directory to be filled with images by 
             means of CTrigonometricMomentImageStack.Store(). Alternatively, this 
             can be a list of file names.

      \param KeywordArguments The following keyword arguments exist:

      \param PhaseSet If this is a set of phase values in radians, all processing 
             steps consider solely images corresponding to phase shifts in this 
             set. However, the four taps belonging to a single phase shift are still 
             considered. By default all phase shifts available in the measurement 
             are used.
      \param Exposure The exposure time of the measurements which are to be used in 
             seconds. Pass None to combine all exposures (default). Note that they 
             are combined in a very naive fashion.
      \param MeasurementFilterFunction A function like that passed to 
             CMeasurement.AddCaptureData() that can be used to define which files 
             should be loaded from the measurement. It works on top of the other 
             corresponding options. Pass None to avoid additional filtering 
             (default).
      \param CalibrationFilterFunction Like MeasurementFilterFunction but for the 
             calibration measurement.
      \param SubtractImages If this parameter is True, images with a relative phase 
             offset of pi are not handled separately but subtracted from one another 
             immediately (default True).
      \param IgnoreTaps If this parameter is True, all procedures will only use the 
             first phase tap stored in each file (default False).
      \param FractionToKeep For measurements where whole images are outliers it can 
             be necessary to remove some takes. This parameter defines the fraction 
             of images to keep (per frequency, phase, tap and exposure). Pass None 
             or 1.0 to keep all data (default).

      \param FinalSmoothing The standard deviation in pixels of the Gaussian blur 
             applied to the calibrated moment image right before biasing (default 
             0.0, i.e. no smoothing).
      \param BiasingThreshold The minimal value of an Eigenvalue of a normalized 
             Toeplitz matrix (default 0.003). Intuitively, greater values employ 
             more biasing to enforce smoother impulse responses. Pass a negative 
             value to disable biasing entirely.
      \param BiasingMode Forwards to CTrigonometricMomentImageStack.ApplyBiasing().

      \param EnablePlotting If this is True (default) various matplotlib plots will 
             be generated to ease analysis of the output.
      \param ShowMomentImages If this is True, all measured images are plotted 
             (default False). This can be time-consuming.
      \param ShowScatterPlots,ShowGraphPlots,ShowFourierPlots,ShowExposurePlots 
             Forwards to CMeasurement.PlotPixelData() (default True,False,False,
             False).
      \param ShowPisarenkoEstimate If this boolean is True (default) the Pisarenko 
             estimate for the final moment images is shown.
      \param PixelDataX,PixelDataY Forwards to CMeasurement.PlotPixelData() as X and 
             Y (default None,None, i.e. a random pixel).

      \param BrightnessFactor The factor by which the magnitude of the moment 
             images is multiplied.

      \return A tuple (Measurement,MomentImageStack) where Measurement is a 
              CMeasurement holding the data loaded from MeasurementDirectory and 
              MomentImageStack is a CTrigonometricMomentImageStack holding the final 
              moment images."""
    EnablePlotting=KeywordArguments.get("EnablePlotting",True);
    # Define an appropriate function for removal of irrelevant measurements
    PhaseSet=KeywordArguments.get("PhaseSet",None);
    Exposure=KeywordArguments.get("Exposure",None);
    AdditionalFilterFunction=KeywordArguments.get("MeasurementFilterFunction",None);
    def FilterFunction(ExposureTime,Frequency,PhaseShift,Take):
        if(not AdditionalFilterFunction is None and not AdditionalFilterFunction(ExposureTime,Frequency,PhaseShift,Take)):
            return False;
        if(np.min(np.abs(np.asarray(FrequencyList)-Frequency))>FrequencyTolerance):
            return False;
        if(PhaseSet!=None and np.min(np.abs(np.exp(1.0j*np.asarray(list(PhaseSet)))-np.exp(1.0j*PhaseShift)))>1.0e-4):
            return False;
        if(Exposure!=None and abs(1.0-Exposure/ExposureTime)>1.0e-4):
            return False;
        return True;
    # Load the measurement
    IgnoreTaps=KeywordArguments.get("IgnoreTaps",False);
    SubtractImages=KeywordArguments.get("SubtractImages",True);
    Measurement=CMeasurement();
    Measurement.AddCaptureData(MeasurementDirectory,FilterFunction,SubtractImages,IgnoreTaps);
    # Print the sorted set of time stamps
    TimeStampList=[MomentImage.TimeStamp for MomentImage in Measurement.MomentImageList];
    print("Moment images were measured at the following times:");
    print(", ".join([str(TimeStamp) for TimeStamp in sorted(list(frozenset(TimeStampList)))]));
    # Visualize the loaded moment images if requested
    if(EnablePlotting and KeywordArguments.get("ShowMomentImages",False)):
        Measurement.ShowMomentImages(FrequencyList);
    print("Loaded "+str(len(Measurement.MomentImageList))+" moment images from "+MeasurementDirectory+".");
    # Get rid of moment images which are outliers
    FractionToKeep=KeywordArguments.get("FractionToKeep",None);
    if(FractionToKeep>=1.0):
        FractionToKeep=None;
    if(not FractionToKeep is None):
        Measurement.FixStaticOutliers(FractionToKeep,EnablePlotting);
    # Visualize the construction of trigonometric moments and the magnitudes of the 
    # images
    if(EnablePlotting):
        Measurement.PlotPixelData(KeywordArguments.get("PixelDataX",None),KeywordArguments.get("PixelDataY",None),FrequencyList,KeywordArguments.get("ShowScatterPlots",True),KeywordArguments.get("ShowGraphPlots",False),KeywordArguments.get("ShowFourierPlots",False),KeywordArguments.get("ShowExposurePlots",False));
        Measurement.PlotValueRanges(FrequencyList);
    # Construct smoothed trigonometric moments
    MomentImageStack=Measurement.GetTrigonometricMomentImageStack(FrequencyList);
    MomentImageStack.Conjugate();
    print("Finished post-processing and plotting for the measurement.");

    # Apply calibration based upon a calibration measurement
    BiasingMode=KeywordArguments.get("BiasingMode","Keep");
    PhaseSet=Measurement.GetPhaseSet();
    CalibrationMeasurement=CMeasurement();
    AdditionalFilterFunction=KeywordArguments.get("CalibrationFilterFunction",None);
    CalibrationMeasurement.AddCaptureData(CalibrationDirectory,FilterFunction,SubtractImages,IgnoreTaps);
    print("Loaded "+str(len(CalibrationMeasurement.MomentImageList))+" moment images from "+CalibrationDirectory+".");
    # Discard phase measurements which are not in the measurement that is to be 
    # calibrated
    for Frequency in FrequencyList:
        CalibrationMeasurement.RemovePhases(Measurement.GetPhaseSet(Frequency),Frequency);
    # Get rid of moment images which are outliers
    if(not FractionToKeep is None):
        CalibrationMeasurement.FixStaticOutliers(0.8,EnablePlotting);
    # Construct trigonometric moments
    CalibrationMomentImageStack=CalibrationMeasurement.GetTrigonometricMomentImageStack(FrequencyList);
    CalibrationMomentImageStack.Conjugate();
    # Compute the calibration factors
    PerPixelScaling=CalibrationMomentImageStack.GetPerPixelCalibration(ReferenceDistribution,BiasingMode=="Discard");
    # Apply the per pixel calibration
    MomentImageStack.ApplyPerPixelCalibration(PerPixelScaling);
    print("Finished calibration.");

    # Apply smoothing to the almost final moment images
    MomentImageStack.Smooth(KeywordArguments.get("FinalSmoothing",0.0));
    # Apply scaling as requested
    if("BrightnessFactor" in KeywordArguments):
        MomentImageStack.ApplyScaling(KeywordArguments["BrightnessFactor"]);
    # Ensure valid moment vectors by means of biasing
    BiasingThreshold=KeywordArguments.get("BiasingThreshold",0.003);
    if(BiasingThreshold>=0.0):
        PixelBiased=MomentImageStack.ApplyBiasing(BiasingThreshold,BiasingMode);
        if(EnablePlotting):
            # Visualize how many pixels have been biased
            Figure=pyplot.figure("Biased areas");
            Figure.suptitle("Biased areas (%.1f%%)"%(100.0*float(np.count_nonzero(PixelBiased))/np.size(PixelBiased)));
            pyplot.imshow(np.where(PixelBiased,1.0,0.0),interpolation="none");
            pyplot.colorbar(ticks=[0.0,1.0]).set_ticklabels(["Unbiased","Biased"]);
    print("Finished post-processing of trigonometric moment images.");

    # Provide output for analysis and further processing
    if(EnablePlotting):
        if(KeywordArguments.get("ShowPisarenkoEstimate",True)):
            MomentImageStack.PlotPisarenkoEstimate();
        MomentImageStack.Display();
    # Store the moment images
    MomentImageStack.Store(OutputDirectory,True,True);
    print("The measurement has been processed.");
    return (Measurement,MomentImageStack);
