# -*- coding: utf-8 -*-
import numpy as np
from scipy import linalg,integrate

## \file
# Provides simplistic implementations of various core methods of the paper.

def GetMaximumEntropySpectralEstimate(TrigonometricMoments,X):
    """This function evaluates a special density function at the given location X 
       (in radians) and returns the result as array of shape identical to that of X. 
       The density function produces the prescribed trigonometric moments (starting 
       with the zeroth moment) and has minimal Burg entropy. If the given moments do 
       not admit a positive density, it returns None."""
    TrigonometricMoments=np.asarray(TrigonometricMoments);
    nFullMoment=TrigonometricMoments.shape[0];
    Toeplitz=linalg.toeplitz(TrigonometricMoments);
    if(np.all(np.linalg.eigh(Toeplitz)[0]>0.0)):
        EvaluationPolynomial=np.linalg.solve(Toeplitz,np.eye(nFullMoment,1));
        return EvaluationPolynomial[0].real/(2.0*np.pi*np.abs(np.polyval(np.conj(EvaluationPolynomial[::-1]),np.exp(1.0j*X)))**2);
    else:
        return None;


def GetMaximumEntropySpectralEstimateCumulative(TrigonometricMoments,X):
    """This function evaluates the cumulative density of the density function given 
       by GetMaximumEntropySpectralEstimate(). The integral is computed numerically 
       using the trapezoidal rule, so X should provide a sufficiently dense 
       sampling. If the given moments do not admit a positive density, it returns 
       None."""
    Density=GetMaximumEntropySpectralEstimate(TrigonometricMoments,X);
    if(not Density is None):
        return integrate.cumtrapz(Density,X,None,0,0.0);
    else:
        return None;


def GetPisarenkoEstimate(TrigonometricMoments):
    """Given a vector of trigonometric moments without (!) a zeroth moment this 
       function computes a distribution consisting of TrigonometricMoments.size 
       Dirac-delta distributions which describes this measurement correctly.
      \return A pair (Phase,Weight) where Phase holds phases in radians (zero to two 
              pi) and Weight holds corresponding weights. Both arrays have the same 
              length as the given vector. Each entry describes one Dirac-delta."""
    # Construct the Toeplitz matrix and extract the eigenvector for the smallest 
    # eigenvalue
    Toeplitz=linalg.toeplitz([0.0]+list(TrigonometricMoments));
    Eigenvalues,Eigenvectors=np.linalg.eigh(Toeplitz);
    Polynomial=Eigenvectors[:,np.argmin(Eigenvalues)];
    # Determine the locations of the Dirac-deltas
    DiracDeltaLocations=np.roots(np.conj(Polynomial[::-1]));
    # Compute the weights via a least squares system
    nMoment=len(TrigonometricMoments);
    Vandermonde=DiracDeltaLocations[np.newaxis,:].repeat(nMoment,0)**np.arange(1,nMoment+1)[:,np.newaxis].repeat(nMoment,1);
    Weights=np.linalg.lstsq(Vandermonde,TrigonometricMoments)[0].real;
    # Construct the output
    Angle=np.angle(DiracDeltaLocations);
    return (np.where(Angle<0.0,Angle+2.0*np.pi,Angle),Weights);


def GetMaximumEntropySpectralEstimateMoment(TrigonometricMoments,MomentIndex):
    """This function computes trigonometric moments of the distribution provided by 
       GetMaximumEntropySpectralEstimate().
      \param TrigonometricMoments The prescribed trigonometric moments of the 
             distribution starting with the zeroth moment.
      \param MomentIndex The integer index of the requested moments. This can be an 
             array of integers in which case the moments for all these indices are 
             computed.
      \return The requested trigonometric moments with shape 
              np.shape(MomentIndex). If the given moments do not admit a 
              distribution with non-vanishing density, None is returned."""
    # Construct the Toeplitz matrix and ensure that it is positive definite
    nMoment=np.shape(TrigonometricMoments)[0]-1;
    Toeplitz=linalg.toeplitz(TrigonometricMoments);
    if(np.any(np.linalg.eigh(Toeplitz)[0]<=0.0)):
        return None;
    # Construct the matrix mapping a moment vector to a moment vector with greater 
    # index
    EvaluationPolynomial=np.linalg.solve(Toeplitz,np.eye(nMoment+1,1).flat);
    MomentRecursionMatrix=np.vstack([
        np.hstack([-EvaluationPolynomial[1:]/EvaluationPolynomial[0],[0.0]]),
        np.hstack([np.eye(nMoment,nMoment),np.zeros((nMoment,1))])]);
    # Compute the required powers of the moment recursion matrix
    Eigenvalues,Eigenvectors=np.linalg.eig(MomentRecursionMatrix);
    IterationStartVector=np.linalg.solve(Eigenvectors,TrigonometricMoments[::-1]);
    Result=np.zeros(np.shape(MomentIndex),dtype=np.complex);
    for i,iMoment in enumerate(np.asarray(MomentIndex).flat):
        if(abs(iMoment)>nMoment):
            PositiveMoment=Eigenvectors[0,:].dot(Eigenvalues**(abs(iMoment)-nMoment)*IterationStartVector);
            Result.flat[i]=PositiveMoment if iMoment>0 else np.conj(PositiveMoment);
        else:
            Result.flat[i]=TrigonometricMoments[iMoment] if iMoment>0 else np.conj(TrigonometricMoments[-iMoment]);
    return Result;


def GetValidMomentCircle(TrigonometricMoments):
    """Given a vector of m trigonometric moments, this function returns the circle 
       containing all acceptable values for trigonometric moment m+1. 
      \return A tuple (Center,Radius) where Center is a complex value such that all 
              complex X fulfilling np.abs(Center-X)<=Radius are valid (m+1)-th 
              trigonometric moments. If the given sequence of moments is invalid by 
              itself, Radius is set to -1.0."""
    TrigonometricMoments=np.asarray(TrigonometricMoments).flatten();
    nMoment=TrigonometricMoments.shape[0]-1;
    Toeplitz=linalg.toeplitz(TrigonometricMoments);
    PreviousDeterminant=np.linalg.det(Toeplitz[0:nMoment,0:nMoment]).real if nMoment!=0 else 1.0;
    Radius=np.linalg.det(Toeplitz).real/PreviousDeterminant;
    CenterMatrix=linalg.toeplitz(np.hstack([TrigonometricMoments,[0.0]]))[1:,0:nMoment+1];
    Center=np.linalg.det(CenterMatrix)*(-1.0)**(nMoment+1)/PreviousDeterminant;
    if(np.all(np.linalg.eigh(Toeplitz)[0]>=0.0)):
        return (Center,Radius);
    else:
        return (Center,-1.0);
