# -*- coding: utf-8 -*-
import numpy as np
from scipy import linalg,optimize
from FiniteDistribution import CFiniteDistribution
from Reconstruction import GetMaximumEntropySpectralEstimate,GetPisarenkoEstimate

## \file
# Implements reconstruction algorithms from the related work. All of them are 
# implemented such that they generate synthetic data from a ground truth and then 
# attempt to reconstruct this ground truth.


def TimeToPhase(Time,Frequency):
    """This function converts a time of flight in seconds to a phase in radians. The 
       frequency has to be given in Hertz.
      \sa PhaseToTime() """
    return 2.0*np.pi*Frequency*Time;

def PhaseToTime(Phase,Frequency):
    """This function converts a phase in radians to a time of flight in seconds. The 
       frequency has to be given in Hertz.
      \sa TimeToPhase() """
    return (1.0/(2.0*np.pi*Frequency))*Phase;

def GetPositiveArgument(NonZeroScalar):
    """Given an array of non-zero, complex scalars this function returns their 
       argument in radians between zero and two pi."""
    Angle=np.angle(NonZeroScalar);
    return np.where(Angle<0.0,Angle+2.0*np.pi,Angle);

def AddNoNoise(X):
    """This function returns the given argument unchanged. It can be passed as 
       functor to techniques to test them with unperturbed input."""
    return X;

def AddGaussianNoise(X,SignalToNoiseRatio):
    """This function contaminates the given vector of positive, real values with 
       noise from a Gaussian normal distribution. The noise is independent for the 
       various scalars. The returned signal has the given signal to noise ratio, 
       i.e. the input is multiplied by a Gaussian with mean one and standard 
       deviation 1.0/SignalToNoiseRatio.
      \return A copy of X contaminated with noise."""
    return np.asarray(X)*np.random.normal(1.0,1.0/SignalToNoiseRatio,np.shape(X));

def GetNoisySinusoidalMeasurements(Distribution,BaseFrequency,FrequencyList,AddNoise):
    """This function returns synthetic measurements with sinusoidal modulation (as 
       complex phasor) and noise as specified.
      \param Distribution The ground truth as CFiniteDistribution parameterized over 
             time of flight in seconds.
      \param BaseFrequency The modulation frequency implied by iFrequencyBegin==1 in 
             hertz.
      \param FrequencyList A list giving the used multiples of the base frequency.
      \param AddNoise A function that can be applied to an array of positive, real 
             scalars holding brightness values measured by the virtual sensor to 
             perturbate it. The perturbed array is returned.
      \return An array of iFrequencyEnd-iFrequencyBegin entries where the i-th entry 
              provides a noisy correlation between the given distribution and 
              np.exp(1.0j*(iFrequencyBegin+i)*Phase)."""
    # Generate the measurement using simulated four-bucket sampling
    BucketMeasurement=np.zeros((len(FrequencyList),4));
    for i,iFrequency in enumerate(FrequencyList):
        for j in range(4):
            BucketMeasurement[i,j]=Distribution.Integrate(lambda Time: 0.5+0.5*np.cos(iFrequency*TimeToPhase(Time,BaseFrequency)-j*0.5*np.pi));
    # Apply noise
    BucketMeasurement=np.maximum(0.0,AddNoise(BucketMeasurement));
    # Generate phasors
    return (BucketMeasurement[:,0]-BucketMeasurement[:,2])+1.0j*(BucketMeasurement[:,1]-BucketMeasurement[:,3]);

def RootsToDistribution(DiracDeltaLocations,BaseFrequency,iFrequencyBegin,Measurements):
    """This function generates a real distribution (which is not guaranteed to have 
       positive weights) describing the given measurements optimally while only 
       having support at prescribed locations.
      \param DiracDeltaLocations The admitted locations of points of support of the 
             resulting distribution as complex values, i.e. the phases of the actual 
             points of support are the positive arguments of these complex values 
             and the magnitude is ignored.
      \param BaseFrequency The base frequency used to define phase.
      \param iFrequencyBegin The beginning of the range of frequencies found in 
             Measurements. The end is implicitly defined by the shape of 
             Measurements.
      \param Measurements The (possibly noisy) measurements for the specified 
             frequencies. See GetNoisySinusoidalMeasurements() for details.
      \return A CFiniteDistribution with support at the prescribed locations 
              expressed as time of flight in seconds and with weights minimizing the 
              least-squares error with respect to the given measurements. The 
              weights may be negative."""
    # Compute the weights via a least-squares system
    DiracDeltaLocations=DiracDeltaLocations/np.abs(DiracDeltaLocations);
    nFrequency=Measurements.size;
    nDiracDelta=DiracDeltaLocations.size;
    Vandermonde=DiracDeltaLocations[np.newaxis,:].repeat(nFrequency,0)**np.arange(iFrequencyBegin,iFrequencyBegin+nFrequency)[:,np.newaxis].repeat(nDiracDelta,1);
    DiracDeltaWeights=np.linalg.lstsq(np.vstack([Vandermonde.real,Vandermonde.imag]),np.hstack([Measurements.real,Measurements.imag]))[0];
    # Construct the result
    Time=PhaseToTime(GetPositiveArgument(DiracDeltaLocations),BaseFrequency);
    return CFiniteDistribution(Time,DiracDeltaWeights,False);


def SPUMIC(Distribution,BaseFrequency,iFrequencyBegin,iFrequencyEnd,AddNoise,nDiracDelta=2):
    """Implements "SPUMIC: Simultaneous Phase Unwrapping and Multipath Interference 
       Cancellation in Time-of-Flight Cameras Using Spectral Methods" by Kirmani et 
       al.
      \param Distribution,BaseFrequency,iFrequencyBegin,iFrequencyEnd,AddNoise 
             Forward to GetNoisySinusoidalMeasurements(). FrequencyList is 
             range(iFrequencyBegin,iFrequencyEnd).
      \param nDiracDelta The number of Dirac-delta pulses to be reconstructed. In 
             the paper the technique is only formulated for two Dirac-deltas so 
             no other values should be used for evaluation of the technique.
      \return A CFiniteDistribution providing the reconstruction. It consists of the 
              requested number of Dirac-deltas and is parameterized over time in 
              seconds."""
    Measurements=GetNoisySinusoidalMeasurements(Distribution,BaseFrequency,range(iFrequencyBegin,iFrequencyEnd),AddNoise);
    # Compute the polynomial for reconstruction
    nFrequency=iFrequencyEnd-iFrequencyBegin;
    HankelMatrix=linalg.hankel(Measurements[0:nFrequency-nDiracDelta],Measurements[nFrequency-nDiracDelta-1:]);
    Polynomial=np.linalg.svd(HankelMatrix)[2][nDiracDelta,:];
    # Solve the polynomial equation to get the location of the Dirac deltas on the 
    # unit circle
    DiracDeltaLocations=np.conj(np.roots(Polynomial[::-1]));
    return RootsToDistribution(DiracDeltaLocations,BaseFrequency,iFrequencyBegin,Measurements);

def BhandariSpectralEstimate(Distribution,BaseFrequency,iFrequencyBegin,iFrequencyEnd,AddNoise):
    """Implements "Resolving Multipath Interference in Kinect: An Inverse Problem 
       Approach" by Bhandari et al.
      \param Distribution,BaseFrequency,iFrequencyBegin,iFrequencyEnd,AddNoise 
             Forward to GetNoisySinusoidalMeasurements(). FrequencyList is 
             range(iFrequencyBegin,iFrequencyEnd) and must have an odd number of 
             entries.
      \return A CFiniteDistribution providing the reconstruction. It consists of 
              (iFrequencyEnd-iFrequencyBegin-1)/2 Dirac-deltas and is parameterized 
              over time in seconds.
      \warning This implementation of the method is rather naive. It is very 
               sensitive to noise. Probably this could be improved by using matrix 
               pencils for inversion of the Vandermonde system as suggested in the 
               paper."""
    if((iFrequencyEnd-iFrequencyBegin)%2==0):
        raise ValueError("Bhandari's spectral estimate requires an odd number of frequency measurements.");
    nDiracDelta=(iFrequencyEnd-iFrequencyBegin-1)//2;
    Measurements=GetNoisySinusoidalMeasurements(Distribution,BaseFrequency,range(iFrequencyBegin,iFrequencyEnd),AddNoise);
    # Construct the Toeplitz matrix for reconstruction
    ToeplitzMatrix=linalg.toeplitz(Measurements[nDiracDelta:],Measurements[nDiracDelta::-1]);
    # Compute the singular vector for the smallest singular value as approximation 
    # to a vector in the (possibly non-existant) null-space
    Polynomial=np.linalg.svd(ToeplitzMatrix)[2][nDiracDelta,:];
    return RootsToDistribution(np.roots(Polynomial[::-1]),BaseFrequency,iFrequencyBegin,Measurements);

def SRA(Distribution,BaseFrequency,FrequencyList,AddNoise,nSample=1000,CovarianceMatrix=None,ErrorTolerance=0.05):
    """Implements "SRA: Fast Removal of General Multipath for ToF Sensors" by 
       Freedman et al.
      \param Distribution,BaseFrequency,FrequencyList,AddNoise See 
             GetNoisySinusoidalMeasurements().
      \param nSample The number of sample points to use for optimization of the 
             impulse response.
      \param CovarianceMatrix If given this has to be a Hermitian positive-definite 
             matrix inducing the norm that is used to measure deviations from a 
             perfect reconstruction in the data term. It must be real with twice as 
             many rows and columns as there are entries in FrequencyList. The first 
             half refers to cosine measurements, the second half to sine 
             measurements. If no matrix is given, the identity matrix is used.
      \param ErrorTolerance The scalar governing the relative error tolerance of the 
             data term.
      \return A CFiniteDistribution with nSample points of support providing the 
              reconstruction parameterized over time in seconds."""
    # Generate the measurements
    Measurements=GetNoisySinusoidalMeasurements(Distribution,BaseFrequency,FrequencyList,AddNoise);
    Measurements=np.hstack([Measurements.real,Measurements.imag]);
    # Generate the matrix of binary strings
    nFrequency=np.size(FrequencyList);
    BinaryStringMatrix=np.zeros((2**(2*nFrequency),2*nFrequency));
    for i in range(BinaryStringMatrix.shape[0]):
        for j in range(BinaryStringMatrix.shape[1]):
            BinaryStringMatrix[i,j]=1.0 if (i&(2**j))!=0 else -1.0;
    # Generate a matrix describing the relation between the unknown density and the 
    # data term
    Phase=np.linspace(0.0,2.0*np.pi,nSample+1)[0:nSample];
    ConstraintMatrix=np.zeros((nFrequency,nSample),dtype=np.complex);
    for i,Frequency in enumerate(FrequencyList):
        ConstraintMatrix[i,:]=np.exp(1.0j*Frequency*Phase);
    ConstraintMatrix=np.vstack([ConstraintMatrix.real,ConstraintMatrix.imag]);
    # Generate the covariance matrix (if necessary) and its inverse square root 
    # (with positive eigenvalues)
    if(CovarianceMatrix is None):
        CovarianceMatrix=np.eye(2*nFrequency,2*nFrequency);
    Eigenvalues,Eigenvectors=np.linalg.eigh(CovarianceMatrix);
    InverseSquareRootMatrix=Eigenvectors.dot(np.diag(1.0/np.sqrt(Eigenvalues))).dot(Eigenvectors.T);
    # Construct the linear programming problem
    LinearInequalityMatrix=BinaryStringMatrix.dot(InverseSquareRootMatrix).dot(ConstraintMatrix);
    LinearInequalityRHS=BinaryStringMatrix.dot(InverseSquareRootMatrix.dot(Measurements));
    LinearInequalityRHS+=ErrorTolerance*np.linalg.norm(Measurements,1);
    # Solve the linear programming problem
    OptimizationResult=optimize.linprog(np.ones((nSample,)),LinearInequalityMatrix,LinearInequalityRHS.flatten());
    if(OptimizationResult.success):
        return CFiniteDistribution(PhaseToTime(Phase,BaseFrequency),OptimizationResult.x,False);
    else:
        raise RuntimeError("Failed to solve the linear programming problem for SRA.");

def ClosedFormInverse(Distribution,BaseFrequency,iFrequencyBegin,AddNoise):
    """Implements the technique with the Dirac-delta-based reconstruction in 
       "Closed-form Inverses for the Mixed Pixel/Multipath Interference Problem in 
       AMCW Lidar" by Godbaz et al.
      \param Distribution,BaseFrequency,iFrequencyBegin,AddNoise Forward to 
             GetNoisySinusoidalMeasurements(). FrequencyList is 
             range(iFrequencyBegin,iFrequencyBegin+4).
      \return The reconstruction which consists of two Dirac-delta distributions 
              parameterized over time in seconds."""
    Measurements=GetNoisySinusoidalMeasurements(Distribution,BaseFrequency,range(iFrequencyBegin,iFrequencyBegin+4),AddNoise);
    # Construct and solve the quadratic equation
    Polynomial=[Measurements[1]*Measurements[3]-Measurements[2]**2,
                Measurements[1]*Measurements[2]-Measurements[0]*Measurements[3],
                Measurements[0]*Measurements[2]-Measurements[1]**2];
    return RootsToDistribution(np.roots(Polynomial[::-1]),BaseFrequency,iFrequencyBegin,Measurements);

def MaximumEntropySpectralEstimate(Distribution,BaseFrequency,nMoment,AddNoise,PhaseSampling=np.linspace(0.0,2.0*np.pi,10001),MinNormalizedEigenvalue=4.0e-3):
    """Implements our proposed method for continuous reconstruction, i.e. the 
       maximum entropy spectral estimate going back to John Parker Burg.
      \param Distribution,BaseFrequency,nMoment,AddNoise Forward to 
             GetNoisySinusoidalMeasurements(). FrequencyList is range(nMoment+1), 
             i.e. nMoment does not count the zeroth moment.
      \param PhaseSampling A sampling of the interval from zero to two pi defining 
             the locations where the returned distribution should be sampled. The 
             returned distribution uses these values converted to time with respect 
             to the given base frequency (see PhaseToTime()). If this sampling is 
             not dense enough, this can lead to severe distortions of cumulative 
             density for nearly sparse distributions.
      \param MinNormalizedEigenvalue If necessary, the zeroth moment is biased to 
             ensure that the smallest eigenvalue of the Toeplitz matrix divided by 
             the zeroth moment has this value.
      \return A CFiniteDistribution with values defined by 
              PhaseToTime(PhaseSampling,BaseFrequency) sampling the density with 
              minimal Burg entropy and prescribed trigonometric moments. If the 
              Pisarenko estimate is used, the distribution only has nMoment freely 
              placed points of support."""
    Measurements=GetNoisySinusoidalMeasurements(Distribution,BaseFrequency,range(nMoment+1),AddNoise);
    Measurements[0]=np.abs(Measurements[0]);
    # Bias the zeroth moment
    MinEigenvalue=np.linalg.eigh(linalg.toeplitz(Measurements))[0].min();
    Measurements[0]+=np.maximum(0.0,(MinEigenvalue-MinNormalizedEigenvalue*Measurements[0])/(MinNormalizedEigenvalue-1.0));
    # Do the reconstruction
    PhaseSampling=np.sort(PhaseSampling);
    Weight=GetMaximumEntropySpectralEstimate(Measurements,PhaseSampling);
    DeltaPhase=PhaseSampling-np.roll(PhaseSampling,1);
    DeltaPhase[0]+=2.0*np.pi;
    return CFiniteDistribution(PhaseToTime(PhaseSampling,BaseFrequency),Weight*DeltaPhase,False);

def PisarenkoEstimate(Distribution,BaseFrequency,nMoment,AddNoise):
    """Implements our proposed method for sparse reconstruction, i.e. the Pisarenko 
       estimate.
      \param Distribution,BaseFrequency,nMoment,AddNoise Forward to 
             GetNoisySinusoidalMeasurements(). iFrequencyBegin is one, iFrequencyEnd 
             is nMoment+1, i.e. nMoment does not count the zeroth moment.
      \return A CFiniteDistribution with nMoment freely placed points of support 
              parameterized over time in seconds."""
    Measurements=GetNoisySinusoidalMeasurements(Distribution,BaseFrequency,range(1,nMoment+1),AddNoise);
    Phase,Weight=GetPisarenkoEstimate(Measurements);
    return CFiniteDistribution(PhaseToTime(Phase,BaseFrequency),Weight,False);
