# -*- coding: utf-8 -*-
from os import path
import numpy as np
from matplotlib import pyplot,animation,cm
from InputOutput import StoreScalarImage
from FiniteDistribution import CFiniteDistribution
import TrigonometricMomentImageStack

class CTransientImage:
    """This class represents a greyscale transient image, i.e. a time-dependent 
       series of common greyscale images showing how a light pulse is reflected by a 
       scene at a high temporal resolution."""
    
    def __init__(self,Image,Phase=None):
        """Initializes this transient image with the given 3D array.
          \param Image See the attribute documentation for details.
          \param Phase A one-dimensional array providing the phase for each frame of 
                 the provided transient image. If this is None, it is assumed that 
                 the transient image uses an equidistant sampling of phases from 0 
                 to 2*np.pi."""
        self.Image=np.asarray(Image);
        if(Phase is None):
            self.Phase=np.linspace(0.0,2.0*np.pi,self.Image.shape[0]+1)[0:self.Image.shape[0]];
        else:
            self.Phase=Phase;

    def Display(self,BrightnessFactor=10.0):
        """Displays this transient image as looping animation with constant frame-
           rate of 30 Hz (i.e. self.Time does not effect playback speed).
          \param BrightnessFactor This factor is multiplied onto the brightness. A 
                 value of 1.0 means that the maximal held pixel value corresponds to 
                 maximal display brightness.
          \return The pyplot.animation.FuncAnimation object used for the animation. 
                  The calling side has to store it!"""
        Figure=pyplot.figure("Transient image");
        Figure.suptitle("Transient image");
        AxesImage=pyplot.imshow(self.Image[0,:,:],cmap=cm.get_cmap("gray"),vmin=0.0,vmax=np.max(self.Image)/BrightnessFactor);
        def UpdateFunction(Frame,TransientImage,AxesImage):
            AxesImage.set_data(TransientImage.Image[Frame,:,:]);
        return animation.FuncAnimation(Figure,UpdateFunction,self.Image.shape[0],lambda:None,(self,AxesImage),interval=33);
    
    def ApplyScaling(self,BrightnessFactor):
        """Multiplies all frames of this transient image by the given scalar 
           factor."""
        self.Image*=BrightnessFactor;
    
    def GetFrameCount(self):
        """Returns the number of frames in this transient image, i.e. the number of 
           available samples along the time axis."""
        return self.Image.shape[0];
    
    def GetDistribution(self,X,Y):
        """This function retrieves a CFiniteDistribution representing the pixel at 
           the given location. The values of the distribution are phases."""
        return CFiniteDistribution(self.Phase,self.Image[:,Y,X],False);
    
    def GetMomentImage(self,MomentGeneratingFunction):
        """This function constructs a moment image from this transient image using a 
           user-defined moment-generating function taking phases in radians. It 
           evaluates the moment-generating function at all points in self.Phase and 
           uses the resulting coefficients to construct a linear combination of the 
           frames. The resulting two-dimensional image has resolution identical to 
           the transient image and every pixel provides a moment for the distribution 
           corresponding to this pixel."""
        Coefficient=MomentGeneratingFunction(self.Phase);
        try:
            return np.tensordot(Coefficient,self.Image,(0,0));
        except MemoryError:
            print("Ran out of memory during generation of a moment image. Falling back to a loop.");
            Result=np.zeros(self.Image.shape[1:3],Coefficient.dtype);
            for i,Factor in enumerate(Coefficient):
                Result+=Factor*self.Image[i,:,:];
            return Result;
    
    def GetTrigonometricMomentImageStack(self,nMoment):
        """This function generates the given number of trigonometric moment images 
           (not counting the zeroth moment) and returns them as 
           CTrigonometricMomentImageStack."""
        return TrigonometricMomentImageStack.CTrigonometricMomentImageStack([self.GetMomentImage(lambda X: np.exp(1.0j*i*X)) for i in range(nMoment+1)]);

    def StoreTransientImage(self,DirectoryPath,Cumulative=False,Normalized=False,StepSize=1):
        """This function stores the raw data of this transient image as sequence of 
           single precision *.exr images. The directory at the given path is filled 
           with images called <Index>.exr providing the incoming radiance at phase 
           self.Phase[Index]. The images are RGB images, but only the red channel 
           contains information.
          \param DirectoryPath The path to the output directory as string.
          \param Cumulative If this parameter is true, the function stores 
                 cumulative brightness rather than momentary brightness, i.e. it 
                 sums up all frames up to the current frame and stores that.
          \param Normalized If this parameter is true, all output pixel values are 
                 divided by the sum of all values for this pixel before they are 
                 output.
          \param StepSize Among StepSize consecutive frames only the first one is 
                 actually written. This does not effect integration of cumulative 
                 transient images."""
        NormalizationFactor=np.ones(self.Image.shape[1:3]);
        if(Normalized):
            ImageSum=self.Image.sum(0);
            NormalizationFactor=1.0/np.where(ImageSum==0.0,1.0,ImageSum);
        elif(Cumulative):
            NormalizationFactor*=1.0/self.Image.shape[0];
        TransientImage=np.zeros_like(self.Image[0,:,:],dtype=np.float);
        for i in range(self.GetFrameCount()):
            if(i%StepSize==0):
                if(not Cumulative):
                    TransientImage=self.Image[i,:,:]*self.Image.shape[0]*NormalizationFactor;
                OutputFileName=path.join(DirectoryPath,str(int(i/StepSize))+".exr");
                StoreScalarImage(TransientImage,OutputFileName);
            if(Cumulative):
                TransientImage+=self.Image[i,:,:]*NormalizationFactor;

    ## A three-dimensional np.ndarray where the first index identifies the time-
    # interval in which light reached the sensor and the other two indices 
    # correspond to height and width, respectively. The values are proportional to 
    # the received radiant energy.
    Image=None;
    ## A one-dimensional array providing the phase of the beginning of the time-
    # interval for each frame in this transient image
    Phase=None;
