# -*- coding: utf-8 -*-
from os import path
import numpy as np
from scipy import ndimage
from matplotlib import pyplot,cm
from InputOutput import LoadScalarImage,StoreScalarImage
import TransientImage
from FiniteDistribution import CFiniteDistribution
from Plotting import GetSubplotLayout,ShowComplexFunction,LinearToSRGB

def GetRootsPerPixel(PolynomialImage):
    """Given an image of shape (Height,Width,Degree+1) this function solves the 
       polynomial equation 
       x**0*PolynomialImage[X,Y,0]+...+x**Degree*PolynomialImage[X,Y,Degree] per 
       pixel and returns all roots.
      \return A complex array with shape (Height,Width,Degree) providing all roots 
              in arbitrary order per pixel."""
    Height,Width,Degree=PolynomialImage.shape;
    Degree-=1;
    # Compute the roots of the polynomials via the companion matrix
    CompanionMatrixImage=np.zeros((Height,Width,Degree,Degree),dtype=np.complex);
    for j in range(1,Degree):
        CompanionMatrixImage[:,:,j,j-1]=1.0;
    NormalizationImage=PolynomialImage[:,:,Degree];
    NormalizationImage=1.0/np.where(NormalizationImage==0.0,1.0,NormalizationImage);
    for j in range(0,Degree):
        CompanionMatrixImage[:,:,j,Degree-1]=-PolynomialImage[:,:,j]*NormalizationImage;
    return np.linalg.eigvals(CompanionMatrixImage);


class CTrigonometricMomentImageStack:
    """This class encapsulates a stack of images containing trigonometric moments. 
       It provides a lot of functionality to analyze and process these images."""
    
    def __init__(self,MomentImageList,Normalized=False):
        """Initializes this class with the given list of images. MomentImageList[j] 
           is supposed to be a two-dimensional array of complex values containing 
           j-th trigonometric moments. For the zeroth trigonometric moment image 
           only the magnitude is considered. Alternatively, MomentImageList can 
           contain file names of *.exr files as produced by StoreScalarImage(). If 
           Normalized is True all moment images except for the zeroth one are 
           multiplied by the zeroth one per pixel."""
        self.MomentImageList=list();
        for MomentImage in MomentImageList:
            if(isinstance(MomentImage,str)):
                self.MomentImageList.append(LoadScalarImage(MomentImage));
            else:
                self.MomentImageList.append(np.asarray(MomentImage));
        self.MomentImageList[0]=np.abs(self.MomentImageList[0]);
        if(Normalized):
            for i in range(1,self.GetFullMomentCount()):
                self.MomentImageList[i]*=self.MomentImageList[0];
        
    def Crop(self,CropOffLeft,CropOffRight,CropOffTop,CropOffBottom):
        """Crops off the specified number of pixels from each of the boundaries of 
           the moment images contained in this stack."""
        for i in range(len(self.MomentImageList)):
            MomentImage=self.MomentImageList[i];
            self.MomentImageList[i]=MomentImage[CropOffTop:MomentImage.shape[0]-CropOffBottom,CropOffLeft:MomentImage.shape[1]-CropOffRight];
            
    def Conjugate(self):
        """This function applies convex conjugation to all held moments (in 
           place)."""
        for j in range(1,self.GetFullMomentCount()):
            self.MomentImageList[j]=np.conj(self.MomentImageList[j]);
    
    def GetPerPixelCalibration(self,ReferenceDistribution=CFiniteDistribution([0.0]),DiscardZerothMomentImage=True):
        """This function computes calibration coefficients, which turn this moment 
           image stack into an image stack describing the given distribution at all 
           pixels (by default a Dirac-delta distribution at zero). The overall 
           brightness (i.e. zeroth moment) remains unchanged unless the zeroth 
           moment is not available or DiscardZerothMomentImage is True. If the given 
           distribution matches the ground truth distribution (at least with respect 
           to its trigonometric moments), it is safe to transfer the resulting per 
           pixel calibration to other measurements.
          \param ReferenceDistribution The reference distribution as 
                 CFiniteDistribution(). In the end only its trigonometric moments 
                 matter. Values should be phases of the base frequency.
          \return Per pixel calibration coefficients in the form expected by 
                  self.ApplyPerPixelCalibration()."""
        # Get trigonometric moments of the reference distribution
        Moments=np.zeros((self.GetFullMomentCount(),),dtype=np.complex);
        for i in range(self.GetFullMomentCount()):
            Moments[i]=ReferenceDistribution.Integrate(lambda Phase: np.exp(1.0j*i*Phase));
        # Compute the moment images that would result from the reference 
        # distribution and take the quotient to obtain calibration factors 
        PerPixelCalibration=np.zeros((self.GetFullMomentCount(),self.GetHeight(),self.GetWidth()),dtype=np.complex);
        DiscardZerothMomentImage=(DiscardZerothMomentImage or np.all(self.MomentImageList[0]<=0.0));
        if(DiscardZerothMomentImage):
            ReferenceBrightness=np.mean(np.abs(self.MomentImageList[1:]),0);
        else:
            ReferenceBrightness=self.MomentImageList[0];
        for i,MomentImage in enumerate(self.MomentImageList):
            ReferenceMomentImage=ReferenceBrightness*Moments[i];
            PerPixelCalibration[i,:,:]=ReferenceMomentImage/MomentImage;
        return np.asarray(PerPixelCalibration);
    
    def ApplyPerPixelCalibration(self,PerPixelScaling):
        """This function applies previously computed calibration values to this 
           stack. PerPixelScaling[j,:,:] is supposed to be an image of scaling 
           factors to be applied to self.MomentImageList[j] as factor."""
        for j in range(self.GetFullMomentCount()):
            self.MomentImageList[j]*=(PerPixelScaling[j,:,:].real if j==0 else PerPixelScaling[j,:,:]);

    def ApplyScaling(self,Scaling):
        """Multiplies all moment images by the given scalar factor."""
        for j in range(self.GetFullMomentCount()):
            self.MomentImageList[j]*=(Scaling.real if j==0 else Scaling);
    
    def ApplyBiasing(self,MinNormalizedEigenvalue=0.01,BiasingMode="Keep"):
        """This function scales values of particular moments wherever this is 
           necessary to make the moment vectors valid. A moment vector is considered 
           valid if the smallest eigenvalue of the corresponding Toeplitz matrix 
           divided by the zeroth moment is greater than or equal to the given 
           threshold.
          \param BiasingMode You can pass four different strings to trigger four 
                 different biasing modes. "Keep" scales all moments except for the 
                 zeroth moment uniformly. "Increase" increases the zeroth moment by 
                 the smallest amount necessary for valid moment vectors. "Discard" 
                 discards the zeroth moment image entirely and instead computes it 
                 to realize the given minimal eigenvalue. This mode is chosen 
                 automatically if zeroth moments are unavailable. "Mask" turns 
                 pixels requiring a bias entirely black.
          \return A boolean image indicating which pixels have been modified."""
        # Enable discarding the zeroth moment image if it is bogus
        if(np.all(self.MomentImageList[0]<=0.0)):
            BiasingMode="Discard";
            print("Automatically enabled discarding the zeroth moment image during biasing because it is all zero.")
        # Compute the minimal eigenvalues per pixel
        ToeplitzArray=self.GetToeplitzImage("ZeroDiagonal");
        MinEigenvalueImage=np.min(np.linalg.eigvalsh(ToeplitzArray).real,2);
        # Generate a top ten of Bogus pixels
        if(BiasingMode!="Discard"):
            MinNormalizedEigenvalueImage=(MinEigenvalueImage/self.MomentImageList[0])[:,:MinEigenvalueImage.shape[1]-3];
            iBogusPixel=np.argsort(MinNormalizedEigenvalueImage.flat)[0:10];
            for i in iBogusPixel:
                Y,X=np.unravel_index(i,MinNormalizedEigenvalueImage.shape);
                Eigenvalue=MinNormalizedEigenvalueImage[Y,X];
                print("Bogus pixel at "+str(X)+", "+str(Y)+" with normalized Eigenvalue "+str(Eigenvalue)+".");
        # Perform the biasing with the appropriate mode
        MinValidZerothMoment=MinEigenvalueImage/(MinNormalizedEigenvalue-1.0);
        Mask=(self.MomentImageList[0]<MinValidZerothMoment);
        if(BiasingMode=="Discard"):
            self.MomentImageList[0]=MinValidZerothMoment;
            return np.ones_like(self.MomentImageList[0],dtype=np.bool8);
        elif(BiasingMode=="Mask"):
            for i in range(len(self.MomentImageList)):
                self.MomentImageList[i]=np.where(Mask,0.0,self.MomentImageList[i]);
            return Mask;
        elif(BiasingMode=="Increase"):
            self.MomentImageList[0]=np.where(Mask,MinValidZerothMoment,self.MomentImageList[0]);
            return Mask;
        else:
            MomentScaling=np.minimum(1.0,(MinNormalizedEigenvalue-1.0)*self.MomentImageList[0]/MinEigenvalueImage);
            for i in range(1,self.GetFullMomentCount()):
                self.MomentImageList[i]*=MomentScaling;
            return MomentScaling!=1.0;
    
    def Smooth(self,StandardDeviationInPixels):
        """Applies smoothing using a Gaussian filter with the specified standard 
           deviation in pixels. The smoothing is applied to all moment images."""
        for i,MomentImage in enumerate(self.MomentImageList):
            if(i==0):
                self.MomentImageList[i]=ndimage.filters.gaussian_filter(MomentImage.real,StandardDeviationInPixels);
            else:
                self.MomentImageList[i]=ndimage.filters.gaussian_filter(MomentImage.real,StandardDeviationInPixels)+1.0j*ndimage.filters.gaussian_filter(MomentImage.imag,StandardDeviationInPixels);

    def PrepareMaximumEntropySpectralEstimate(self,ForceRecomputation=False):
        """This function pre-computes coefficients needed for the maximum entropy 
           spectral estimate which depend upon the held moment images. This way it 
           ensures that GetMaximumEntropySpectralEstimate() can function 
           efficiently. You do not need to call it yourself unless you have changed 
           moment images after using a maximum entropy spectral estimate. In this 
           case you should pass True for ForceRecomputation."""
        if(self.EvaluationPolynomialImage is None or ForceRecomputation):
            # Construct an image with Toeplitz matrices per pixel
            nFullMoment=self.GetFullMomentCount();
            ToeplitzImage=self.GetToeplitzImage("Default");
            # Compute evaluation polynomials per pixel
            RHSImage=np.zeros((self.GetHeight(),self.GetWidth(),nFullMoment),dtype=np.complex);
            RHSImage[:,:,0]=1.0;
            self.EvaluationPolynomialImage=np.nan_to_num(np.linalg.solve(ToeplitzImage,RHSImage));
    
    def GetMomentVector(self,X,Y):
        """This function returns the moment vector for the pixel with the given 
           coordinates as one-dimensional numpy array."""
        return np.asarray([MomentImage[Y,X] for MomentImage in self.MomentImageList]);
    
    def GetMaximumEntropySpectralEstimate(self,Phase):
        """This function returns an image of brightness values for the given phase. 
           The brightness per pixel as function of the phase is a density matching 
           all held trigonometric moments while realizing the minimal possible Burg 
           entropy."""
        if(self.EvaluationPolynomialImage is None):
            self.PrepareMaximumEntropySpectralEstimate();
        FourierSeriesImage=np.tensordot(np.conj(self.EvaluationPolynomialImage),np.exp(1.0j*Phase)**np.asarray(range(self.GetFullMomentCount())),[2,0]);
        return self.EvaluationPolynomialImage[:,:,0].real/(2.0*np.pi*np.abs(FourierSeriesImage)**2);
    
    def GetTransientImage(self,PhaseSampling=np.linspace(0.0,2.0*np.pi,301)[:300]):
        """This function returns a CTransientImage containing the maximum entropy 
           spectral estimate of the transient image described by the held 
           trigonometric moment images.
          \param PhaseSampling An iterable of phases at which the density should be 
                 sampled. This should be a sorted list of equidistant samples."""
        self.PrepareMaximumEntropySpectralEstimate();
        StepWidth=np.mean(PhaseSampling[1:]-PhaseSampling[:PhaseSampling.size-1]);
        return TransientImage.CTransientImage([np.asarray(self.GetMaximumEntropySpectralEstimate(Phase)*StepWidth,dtype=np.float32) for Phase in PhaseSampling],PhaseSampling);
    
    def GetGodbazPhaseImage(self):
        """This function computes a phase image using the technique proposed by 
           Godbaz et al. in the paper "Closed-form Inverses for the Mixed Pixel/
           Multipath Interference Problem in AMCW Lidar".
          \return An image of phases as radians in the range from zero to two pi.
          \pre self.GetMomentCount()==4."""
        if(self.GetMomentCount()!=4):
            raise ValueError("CTrigonometricMomentImageStack.GetGodbazPhaseImage() requires exactly four moments.");
        # Construct and solve the quadratic equation
        Measurements=self.MomentImageList[1:];
        PolynomialImage=np.dstack(
                    [Measurements[1]*Measurements[3]-Measurements[2]**2,
                     Measurements[1]*Measurements[2]-Measurements[0]*Measurements[3],
                     Measurements[0]*Measurements[2]-Measurements[1]**2]
        );
        RootImage=GetRootsPerPixel(PolynomialImage);
        # Estimate the strength of the returns
        VandermondeImage=np.zeros((self.GetHeight(),self.GetWidth(),self.GetMomentCount(),2),dtype=np.complex);
        MomentImage=np.zeros((self.GetHeight(),self.GetWidth(),self.GetMomentCount()),dtype=np.complex);
        for i in range(4):
            VandermondeImage[:,:,i,:]=(RootImage/np.abs(RootImage))**(i+1);
            MomentImage[:,:,i]=self.MomentImageList[i+1];
        GramImage=np.zeros((self.GetHeight(),self.GetWidth(),2,2),dtype=np.complex);
        RHSImage=np.zeros((self.GetHeight(),self.GetWidth(),2),dtype=np.complex);
        for i in range(2):
            for k in range(4):
                for j in range(2):
                    GramImage[:,:,i,j]+=np.conj(VandermondeImage[:,:,k,i])*VandermondeImage[:,:,k,j];
                RHSImage[:,:,i]+=np.conj(VandermondeImage[:,:,k,i])*MomentImage[:,:,k];
        WeightImage=np.linalg.solve(GramImage,RHSImage).real;
        # Select the phase with the stronger return
        RelevantRoot=np.where(WeightImage[:,:,0]>WeightImage[:,:,1],RootImage[:,:,0],RootImage[:,:,1]);
        Result=np.angle(RelevantRoot);
        return np.where(Result<0.0,Result+2.0*np.pi,Result);

    def GetPhaseImage(self):
        """This function computes the phase of the global maximum of the maximum 
           entropy spectral estimate per pixel and returns it as image with radians 
           in the range from zero to two pi."""
        self.PrepareMaximumEntropySpectralEstimate();
        # Construct the derivative of the evaluation polynomial per pixel
        nMoment=self.GetMomentCount();
        DerivativePolynomialImage=np.zeros((self.GetHeight(),self.GetWidth(),nMoment*2+1),dtype=np.complex);
        for j in range(nMoment+1):
            for k in range(nMoment+1):
                DerivativePolynomialImage[:,:,nMoment+j-k]+=np.conj(self.EvaluationPolynomialImage[:,:,j])*self.EvaluationPolynomialImage[:,:,k]*(k-j);
        # Construct the companion matrix of the derivative polynomial per pixel
        RootImage=GetRootsPerPixel(DerivativePolynomialImage);
        # Evaluate the evaluation polynomial at each of these roots
        FourierSeriesImage=np.zeros_like(RootImage);
        for i in range(nMoment+1):
            PolynomialCoefficient=np.tile(self.EvaluationPolynomialImage[:,:,i:i+1],[1,1,RootImage.shape[2]]);
            FourierSeriesImage+=np.conj(PolynomialCoefficient)*RootImage**i;
        # Identify the smallest value of the evaluation polynomial (the density is 
        # the reciprocal) masking out roots which are not on the unit circle
        MaskValue=np.max(np.abs(FourierSeriesImage));
        MaskedMagnitudeImage=np.where(np.abs(np.abs(RootImage)-1.0)<1.0e-4,np.abs(FourierSeriesImage),MaskValue);
        iRootImage=np.argmin(MaskedMagnitudeImage,2);
        # Convert the relevant root to a phase and return it
        OptimalRootImage=np.choose(iRootImage,[RootImage[:,:,i] for i in range(nMoment*2)]);
        Result=np.angle(OptimalRootImage);
        return np.where(Result<0.0,Result+2.0*np.pi,Result);
    
    def GetPisarenkoEstimate(self):
        """This function constructs a Pisarenko estimate for each pixel, i.e. a 
           distribution consisting of exactly self.GetMomentCount() Dirac-delta 
           distributions, which realizes all held moments except for the zeroth one.
          \return A tuple (Phase,Weight) consisting of two real-valued arrays of 
                  shape (self.GetHeight(),self.GetWidth(),self.GetMomentCount()). 
                  Phase contains the locations of Dirac-deltas for each pixel using 
                  radians in the range from zero to two pi, Weight contains the 
                  corresponding weights. The weights are sorted in decreasing 
                  order."""
        # Compute the eigenvector for the minimal eigenvalue of the Toeplitz matrix 
        # per pixel
        nMoment=self.GetMomentCount();
        ToeplitzArray=self.GetToeplitzImage("ZeroDiagonal");
        Eigenvalues,Eigenvectors=np.linalg.eigh(ToeplitzArray);
        MinEigenvectorImage=np.choose(np.argmin(Eigenvalues,2)[:,:,np.newaxis].repeat(nMoment+1,2),[Eigenvectors[:,:,:,i] for i in range(nMoment+1)]);
        # Compute the roots of the induced polynomials
        RootImage=GetRootsPerPixel(np.conj(MinEigenvectorImage));
        # Compute the phases for the Dirac-delta distributions
        PhaseImage=np.angle(RootImage);
        PhaseImage=np.where(PhaseImage<0.0,PhaseImage+2.0*np.pi,PhaseImage);
        # Compute weights by solving a linear system
        VandermondeImage=np.zeros((self.GetHeight(),self.GetWidth(),self.GetMomentCount(),self.GetMomentCount()),dtype=np.complex);
        MomentImage=np.zeros_like(PhaseImage,dtype=np.complex);
        for i in range(nMoment):
            VandermondeImage[:,:,i,:]=np.exp(PhaseImage*1.0j*(i+1));
            MomentImage[:,:,i]=self.MomentImageList[i+1];
        WeightImage=np.linalg.solve(VandermondeImage,MomentImage).real;
        # Sort by decreasing weight
        Permutation=np.argsort(-WeightImage,2);
        XIndex,YIndex=np.meshgrid(range(self.GetWidth()),range(self.GetHeight()));
        PhaseImage=np.dstack([PhaseImage[YIndex,XIndex,Permutation[:,:,j]] for j in range(nMoment)]);
        WeightImage=np.dstack([WeightImage[YIndex,XIndex,Permutation[:,:,j]] for j in range(nMoment)]);
        return (PhaseImage,WeightImage);
    
    def SeparateDirectIllumination(self,MinDirectFraction=0.2):
        """This function uses a Pisarenko estimate to separate the direct component 
           of illumination from indirect components.
          \param MinDirectFraction The minimal fraction of the total weight which 
                 should be contained in the direct component. The direct component 
                 is the component with minimal phase surpassing this threshold. Pass 
                 None for this threshold to define that the strongest component is 
                 the direct component.
          \return A triple (Direct,Indirect,DirectPhase) consisting of three real-
                  valued arrays of shape (self.GetHeight(),self.GetWidth()). Direct 
                  contains the weights of the Dirac-delta describing the direct 
                  component. Indirect is the difference to the zeroth moment image. 
                  Phase contains the phase associated with the direct return."""
        PhaseImage,WeightImage=self.GetPisarenkoEstimate();
        if(not MinDirectFraction is None):
            RelevantPhaseImage=np.where(WeightImage>MinDirectFraction*self.MomentImageList[0][:,:,np.newaxis].repeat(WeightImage.shape[2],2),PhaseImage,2.0*np.pi);
            iDirectReturn=np.argmin(RelevantPhaseImage,2);
        else:
            iDirectReturn=np.argmax(WeightImage,2);
        Direct=np.choose(iDirectReturn,[WeightImage[:,:,i] for i in range(self.GetMomentCount())]);
        DirectPhase=np.choose(iDirectReturn,[PhaseImage[:,:,i] for i in range(self.GetMomentCount())]);
        Indirect=self.MomentImageList[0]-Direct;
        return (Direct,Indirect,DirectPhase);
    
    def GetErrorBound(self,PoissonKernelRadius,Phase):
        """This function computes a lower and upper bound for the value of a density 
           function having the held trigonometric moments per pixel. Since such 
           densities can usually be arbitrary big or small, the bounds actually 
           refer to the density obtained after smoothing with a Poisson kernel.
          \param PoissonKernelRadius A parameter controlling the sharpness of the 
                 kernel. For zero it is uniform, for one it converges to a Dirac-
                 delta at zero. This has to be less than one.
          \param Phase The phase in radians at which the upper and lower bound 
                 should be evaluated. This has to be an array with shape (nSample,).
          \return A pair (LowerBound,UpperBound) of arrays of shape 
                  (self.GetHeight(),self.GetWidth(),nSample) where LowerBound[Y,X,i] 
                  provides the minimal smoothed density at Phase[i] and pixel (X,Y). 
                  UpperBound provides corresponding upper bounds."""
        # Convert the phases to parameters for the Poisson kernel on the unit disc
        ZSamples=PoissonKernelRadius*np.exp(-1.0j*Phase);
        # Invert the Toeplitz matrix per pixel
        ToeplitzImage=self.GetToeplitzImage("Default");
        InvToeplitzImage=np.linalg.inv(ToeplitzImage);
        # Handle different samples separately
        nSample=Phase.shape[0];
        LowerBound=np.zeros([self.GetHeight(),self.GetWidth(),nSample]);
        UpperBound=np.zeros_like(LowerBound);
        for i,Z in enumerate(ZSamples):
            # Precompute all values of the various polynomials per pixel 
            nFullMoment=self.GetFullMomentCount();
            MomentPolynomial=np.zeros(self.MomentImageList[0].shape+(nFullMoment,),dtype=np.complex);
            for j in range(0,nFullMoment):
                MomentPolynomial[:,:,j]=self.MomentImageList[0];
                for k in range(1,j+1):
                    MomentPolynomial[...,j]+=2.0*self.MomentImageList[k]*Z**k;
                MomentPolynomial[...,j]*=Z**(-j-1);
            # Accumulate the individual terms of the rational function per pixel
            LHSNumerator=np.ones(self.MomentImageList[0].shape,dtype=np.complex)*2.0/(1.0-PoissonKernelRadius**2);
            RHSNumerator=np.zeros_like(LHSNumerator);
            Denominator=np.zeros_like(LHSNumerator);
            for j in range(0,nFullMoment):
                for k in range(0,nFullMoment):
                    LHSNumerator+=np.conj(MomentPolynomial[...,j])*InvToeplitzImage[...,j,k]*Z**(-k-1);
                    RHSNumerator+=np.conj(MomentPolynomial[...,j])*InvToeplitzImage[...,j,k]*MomentPolynomial[...,k];
                    Denominator+=np.conj(Z**(-j-1))*InvToeplitzImage[...,j,k]*Z**(-k-1);
            # Compute interval containing all valid densities
            Center=LHSNumerator/Denominator;
            Radius=np.sqrt(np.abs(Center)**2-RHSNumerator/Denominator).real;
            LowerBound[:,:,i]=Center.real-Radius;
            UpperBound[:,:,i]=Center.real+Radius;
        # Return the result accounting for the factor of two pi in the trigonometric 
        # moments
        return (LowerBound*0.5/np.pi,UpperBound*0.5/np.pi);

    def GetToeplitzImage(self,Mode="Default"):
        """This function constructs a Hermitian Toeplitz matrix for each pixel.
          \param Mode Determines the contents of the Toeplitz matrix. "Default" 
                 generates a Toeplitz matrix containing all held moments in the 
                 appropriate diagonals, "Normalized" is like "Default" except that 
                 all moments are divided by the zeroth moment and "ZeroDiagonal" 
                 treats the zeroth moment as zero.
          \return An array of shape (self.GetHeight(),self.GetWidth(),self.GetFullMomentCount(),self.GetFullMomentCount()) 
                  where entry [Y,X,:,:] is the requested Toeplitz matrix for the 
                  pixel at (X,Y)."""
        nFullMoment=self.GetFullMomentCount();
        ToeplitzImage=np.zeros((self.GetHeight(),self.GetWidth(),nFullMoment,nFullMoment),dtype=np.complex);
        if(Mode=="Default"):
            Factor=1.0;
            DiagonalEntry=self.MomentImageList[0];
        elif(Mode=="Normalized"):
            Factor=self.MomentImageList[0];
            Factor=1.0/np.where(Factor==0.0,1.0,Factor);
            DiagonalEntry=1.0;
        elif(Mode=="ZeroDiagonal"):
            Factor=1.0;
            DiagonalEntry=0.0;
        for j in range(nFullMoment):
            for k in range(nFullMoment):
                if(j>k):
                    ToeplitzImage[:,:,j,k]=self.MomentImageList[j-k]*Factor;
                elif(j==k):
                    ToeplitzImage[:,:,j,k]=DiagonalEntry;
                else:
                    ToeplitzImage[:,:,j,k]=np.conj(self.MomentImageList[k-j])*Factor;
        return ToeplitzImage;
    
    def Store(self,DirectoryPathOrFileList,Normalize=True,Print=False):
        """This function stores the held trigonometric moment images to the 
           specified location. The image format is the same as for StoreScalarImage().
          \param DirectoryPathOrFileList If this is a string, it specifies a 
                 directory. The images are stored to this directory using 
                 Measurement<FrequencyIndex>.exr as file name. Otherwise, it has 
                 to be a list of strings specifying full file paths for every single 
                 moment image.
          \param Normalize Pass True to ensure that all moment images (except for 
                 the zeroth one) are divided per pixel by the zeroth one.
          \param Print If this boolean is True, the function prints information 
                 about produced files."""
        if(isinstance(DirectoryPathOrFileList,str)):
            FileList=[path.join(DirectoryPathOrFileList,"MomentImage"+str(i)+".exr") for i in range(self.GetFullMomentCount())];
        else:
            FileList=DirectoryPathOrFileList;
        FirstImage=True;
        for FilePath,MomentImage in zip(FileList,self.MomentImageList):
            # Store the image
            if(Normalize and not FirstImage):
                StoreScalarImage(MomentImage/self.MomentImageList[0],FilePath);
            else:
                StoreScalarImage(MomentImage,FilePath);
            FirstImage=False;
            if(Print):
                print("Wrote a "+str(self.GetWidth())+"*"+str(self.GetHeight())+" moment image to "+FilePath+".");
    
    def GetWidth(self):
        """This function returns the width of the held moment images in pixels."""
        return self.MomentImageList[0].shape[1];
    
    def GetHeight(self):
        """This function returns the height of the held moment images in pixels."""
        return self.MomentImageList[0].shape[0];
    
    def GetMomentCount(self):
        """This function returns the number of trigonometric moments available per 
           pixel _NOT_ counting the zeroth moment."""
        return len(self.MomentImageList)-1;
    
    def GetFullMomentCount(self):
        """This function returns the number of trigonometric moments available per 
           pixel counting the zeroth moment."""
        return len(self.MomentImageList);
    
    def PlotEigenvalues(self,Normalize=True):
        """This method visualizes the eigenvalues of the Toeplitz matrix per pixel.
          \param Normalize Pass True to get Eigenvalues divided by the zeroth 
                 moment."""
        # Compute the eigenvalues per pixel and sort them
        ToeplitzArray=self.GetToeplitzImage("Normalized" if Normalize else "Default");
        EigenvalueImage=np.linalg.eigvalsh(ToeplitzArray).real;
        EigenvalueImage.sort(2);
        # Plot them
        NormalizationString=("" if Normalize else "not ")+"normalized";
        Figure=pyplot.figure("Eigenvalues ("+NormalizationString+", from smallest to largest)",tight_layout=True);
        nRow,nColumn=GetSubplotLayout(self.GetFullMomentCount());
        for i in range(self.GetFullMomentCount()):
            Figure.add_subplot(nRow,nColumn,i+1,xticks=[],yticks=[],title="$\lambda_"+str(i)+"$");
            pyplot.imshow(EigenvalueImage[:,:,i]);
            pyplot.colorbar(orientation="horizontal",pad=0.01);
            
    def PlotPisarenkoEstimate(self):
        """This function plots weight images and phase images resulting from the 
           Pisarenko estimate."""
        Phase,Weight=self.GetPisarenkoEstimate();
        Figure=pyplot.figure("Pisarenko estimate",tight_layout=True);
        for i in range(self.GetMomentCount()):
            Figure.add_subplot(2,self.GetMomentCount(),i+1,xticks=[],yticks=[],title="Weight "+str(i));
            pyplot.imshow(LinearToSRGB(Weight[:,:,i]/np.percentile(Weight,98)),cmap=cm.get_cmap("gray"),vmin=0.0,vmax=1.0);
            Figure.add_subplot(2,self.GetMomentCount(),i+self.GetMomentCount()+1,xticks=[],yticks=[],title="Phase "+str(i));
            pyplot.imshow(Phase[:,:,i],cmap=cm.get_cmap("coolwarm"));
            pyplot.colorbar(orientation="horizontal");
    
    def Display(self):
        """This method displays this stack of moment images in a human-friendly 
           fashion."""
        Figure=pyplot.figure("Moment images",tight_layout=True);
        nPlotRow,nPlotColumn=GetSubplotLayout(self.GetFullMomentCount());
        AverageMagnitude=np.mean(np.abs(np.nan_to_num(np.array(self.MomentImageList))));
        for i,MomentImage in enumerate(self.MomentImageList):
            Figure.add_subplot(nPlotRow,nPlotColumn,i+1,xticks=[],yticks=[],title=r"$\mathbb{E}(\exp(i\cdot"+str(i)+r"\cdot\mathbf{\varphi}))$");
            if(i==0):
                pyplot.imshow(MomentImage,cmap=cm.get_cmap("gray"),vmin=0.0,vmax=np.max(MomentImage));
            else:
                ShowComplexFunction(MomentImage/AverageMagnitude);
    
    ## A list containing all moment images sorted by the index of the held moments
    MomentImageList=list();
    ## This array holds coefficients used by GetMaximumEntropySpectralEstimate(). If 
    # PrepareMaximumEntropySpectralEstimate() has not been invoked yet, it is None.
    EvaluationPolynomialImage=None;
